/**
  *  \file client/dlg-mission.cc
  *  \brief Starship Mission Editor
  *
  *  This module contains all functionality to edit a starship
  *  mission. This includes choosing the mission itself as well as
  *  setting the parameters.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-visual.h"
#include "client/dlg-mission.h"
#include "client/screens.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "game/host.h"
#include "game/global.h"
#include "game/mission.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/wscanchart.h"
#include "game/racename.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/numbersel.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"
#include "game/fleet.h"
#include "client/widgets/objlist.h"

/**************************** WExtendedMission ***************************/

namespace {
    /** Extended Mission Selector Window. This window displays three
        input lines to enter the M.I.T. values. This window can be run
        and returns a boolean value (1=ok, 0=cancel). On OK, places
        input in a TMissionData object. */
    class WExtendedMission : public UIWindow {
        ObjectHolder  h;
        UIButton      ok, cancel;
        UIInputLine   m, i, t;
        TMissionData& output;
        UIGridLayout  gridl;
        UIGroup       gridg;
        void init(int sid);
     public:
        WExtendedMission(TMissionData& output, int sid);
        void onOk();
        void onCancel();
    };
}

/** Constructor.
    \param output   on success, mission data is placed here
    \param sid      if nonzero, must be a played ship id. The window
                    will be initialized with this ship's mission. Otherwise,
                    the window will be initialized with zero. */
WExtendedMission::WExtendedMission(TMissionData& output, int sid)
    : UIWindow(_("Extended Mission"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      ok(0, _("OK"), SDLK_RETURN),
      cancel(0, _("Cancel"), SDLK_ESCAPE),
      m(5, ss_Alt + 'm', UIInputLine::ilf_Framed | UIInputLine::ilf_NumbersOnly, 1000),
      i(5, ss_Alt + 'i', UIInputLine::ilf_Framed | UIInputLine::ilf_NumbersOnly, 1001),
      t(5, ss_Alt + 't', UIInputLine::ilf_Framed | UIInputLine::ilf_NumbersOnly, 1002),
      output(output),
      gridl(2), gridg(gridl)
{
    init(sid);
}

void
WExtendedMission::init(int sid)
{
    setLayout(UIVBoxLayout::instance5);

    gridg.add(h.add(new UIStaticText(_("Mission number:"), UIStaticText::StaticTitle)));
    gridg.add(m);
    gridg.add(h.add(new UIStaticText(_("Intercept number:"), UIStaticText::StaticTitle)));
    gridg.add(i);
    gridg.add(h.add(new UIStaticText(_("Tow number:"), UIStaticText::StaticTitle)));
    gridg.add(t);
    add(gridg);

    UIGroup& bs = h.add(new UIGroup(UIHBoxLayout::instance5));
    bs.add(ok);
    bs.add(cancel);
    bs.add(h.add(new UISpacer()));
    add(bs);
    add(h.add(new UIFocusIterator(1000, 1002, UIFocusIterator::fi_Tab | UIFocusIterator::fi_Vert)));
    pack();

    ok.sig_fired.add(this, &WExtendedMission::onOk);
    cancel.sig_fired.add(this, &WExtendedMission::onCancel);

    /* initialize input fields */
    if (sid) {
        GShip& s = getDisplayedTurn().getCurrentUniverse().getShip(sid);
        m.setText(itoa(s.getMission()));
        i.setText(itoa(s.getInterceptId()));
        t.setText(itoa(s.getTowId()));
    } else {
        m.setText("0");
        i.setText("0");
        t.setText("0");
    }
}

/** "OK" handler. Validates input and does stop(1). */
void
WExtendedMission::onOk()
{
    /* This allows the full 0-10000 range for all parameters even if we thing
       the host has a tighter limit, to allow it to be used as an emergengy
       overide if host changes behaviour. */
    long mm, ii, tt;
    if (!stringValue(m.getText(), mm) || mm < 0 || mm > 10000) {
        m.setFocusedWidget();
        return;
    }
    if (!stringValue(i.getText(), ii) || ii < 0 || ii > 10000) {
        i.setFocusedWidget();
        return;
    }
    if (!stringValue(t.getText(), tt) || tt < 0 || tt > 10000) {
        t.setFocusedWidget();
        return;
    }

    output.mission = mm;
    output.args[0] = ii;
    output.args[1] = tt;

    stop(1);
}

/** Cancel handler. */
void
WExtendedMission::onCancel()
{
    stop(0);
}

/*************************** Ship / Planet Arg ***************************/

namespace {
    class WShipArgWindow : public UIWindow {
        ObjectHolder         h;
        WScannerChartWidget  chart;
        int                  sid, flags, cur;
        int                  result;
        void onOk();
     public:
        WShipArgWindow(int sid, int flags, int cur, string_t title);
        ~WShipArgWindow();

        int getResult() const { return result; }
    };
}

WShipArgWindow::WShipArgWindow(int sid, int flags, int cur, string_t title)
    : UIWindow(title, BLUE_WINDOW, COLOR_GRAY),
      sid(sid), flags(flags), cur(cur), result(0)
{
    setLayout(UIVBoxLayout::instance5);

    if (sid)
        chart.setCenterPosition(getDisplayedTurn().getCurrentUniverse().getShip(sid).getPos());
    else
        chart.setCenterPosition(GPoint(2000, 2000) /* FIXME? */);
    chart.setScannerOrigin(chart.getCenterPosition());

    if (getDisplayedTurn().getCurrentUniverse().ty_any_ships.isValidIndex(cur))
        chart.setScannerTarget(chart.clipScannerPosition(getDisplayedTurn().getCurrentUniverse().getShip(cur).getPos()));
    else if (sid)
        chart.setScannerTarget(chart.clipScannerPosition(getDisplayedTurn().getCurrentUniverse().getShip(sid).getWaypoint()));
    else
        chart.setScannerTarget(chart.getCenterPosition());

    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard | chart.sc_ShipsOnly, true);

    add(chart);

    UIGroup& g    = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIButton& ok  = h.add(new UIButton(0,  _("F10 - OK"),      SDLK_F10));
    UIButton& esc = h.add(new UIButton(0,  _("ESC - Cancel"),  SDLK_ESCAPE));
    esc.addStop();
    ok.sig_fired.add(this, &WShipArgWindow::onOk);

    g.add(h.add(new UISpacer()));
    g.add(ok);
    g.add(esc);
    add(g);

    const int size = ui_root->getExtent().h;
    setExtent(GfxRect(0, 0, size - 50, size - 10));
}

WShipArgWindow::~WShipArgWindow()
{ }

void
WShipArgWindow::onOk()
{
    int lflags = ((flags & GMission::ma_Own) == 0
                  ? ls_AllowForeignShips | ls_IncludeForeignShips
                  : 0);
    int lexcl  = ((flags & GMission::ma_NotThis) == 0 ? 0 : sid);

    GObjectReference chosen =
        doListShips(chart.getScannerTarget(),
                    lflags | ls_EarlyExit, lexcl, getName(),
                    _("OK"));

    if (chosen.isValid()) {
        result = chosen->getId();
        stop(1);
    }
}

/**************** Argument Entry for One-Argument Mission ****************/

static long
getArgLimit(int format)
{
    switch (format & GMission::ma_TypeMask) {
     case GMission::mat_PID:
     case GMission::mat_BID:
        return NUM_PLANETS;
     case GMission::mat_Player:
        return NUM_PLAYERS;
     case GMission::mat_SID:
     case GMission::mat_SID_here:
        return NUM_SHIPS;
     default:
        return host.getCommandArgumentLimit();
    }
}

/** Query argument for one-argument mission.
    \param mission     the mission we're trying to set
    \param output      parameter goes here.
    \param index       parameter slot
    \param sid         ship id (played ship, or zero)
    \return true if user chose OK, false if Cancel. */
static bool
getMissionArg(const GMission& mission, int& output, GMission::Parameter index, int sid)
{
    const int flags = mission.getParameterType(index);
    switch (flags & GMission::ma_TypeMask) {
     case GMission::mat_PID:
        /* planet Id */
        if (flags & GMission::ma_Own) {
            if (int chosen = choosePlanet()) {
                output = chosen;
                return true;
            }
        } else {
            /* maybe we can also use a better interface
               FIXME: we should also place the cursor on the right
               position. Low-priority, though; no known mission uses
               planet args. */
            GObjectList list;
            for (int i = current_anyplanet_type.findNextIndex(0); i != 0; i = current_anyplanet_type.findNextIndex(i)) {
                list.addObject(current_anyplanet_type, i);
            }
            list.postprocessUser(true);

            GObjectReference chosen = WObjectList(0, 20, 320, list).doStandardDialog2(mission.getName(), "");
            if (chosen.isValid()) {
                output = chosen->getId();
                return true;
            }
        }
        return false;
     case GMission::mat_SID: {
        /* ship Id */
        WShipArgWindow win(sid, flags, output, mission.getName());
        ui_root->centerWidget(win);
        ui_root->add(win);
        int rv = ui_root->run();
        ui_root->close(win);
        if (rv) {
            output = win.getResult();
            return true;
        }
        return false;
     }
     case GMission::mat_SID_here: {
        /* Id of a ship here */
        if (!sid)
            return false;       // cannot happen, but you never know

        int flags;
        if (mission.getParameterType(index) & GMission::ma_Own)
            flags = ls_EarlyExit;
        else
            flags = ls_EarlyExit | ls_IncludeForeignShips | ls_AllowForeignShips;

        GObjectReference chosen =
            doListShips(getDisplayedTurn().getCurrentUniverse().getShip(sid).getPos(),
                        flags,
                        (mission.getParameterType(index) & GMission::ma_NotThis) ? sid : 0,
                        mission.getName(),
                        _("OK"));
        if (chosen.isValid()) {
            output = chosen->getId();
            return true;
        }
        return false;
     }
     case GMission::mat_BID:
        /* base Id */
        if (int chosen = chooseBase()) {
            output = chosen;
            return true;
        }
        return false;
     case GMission::mat_Player: {
        /* player Id */
        UIStringList list;
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            if ((flags & GMission::ma_NotThis) == 0 || sid == 0 || i != getDisplayedTurn().getCurrentUniverse().getShip(sid).getOwner())
                list.add(i, format("%X - %s", i, player_racenames.getShortName(i)));

        UIStandardListbox listbox(0, -1, 240, list);
        listbox.scrollToKey(output);
        if (listbox.doStandardListbox(mission.getName(), 0)) {
            output = listbox.getKey();
            return true;
        }
        return false;
     }
     default: {
        /* number */
        UIDecimalSelector arg(0, 0, getArgLimit(flags), 10);
        arg.setValue(output);
        if (arg.doStandardDialog(mission.getName(), mission.getParameterName(index) + ":", 0)) {
            output = arg.getValue();
            return true;
        }
        return false;
     }
    }
}

/**************** Argument Entry for Two-Argument Mission ****************/

namespace {
    /** Argument display/edit widget for one argument. */
    class WShipArgWidget : public UINumberSelector {
        const GMission& mission;
        GMission::Parameter which;
        int sid;

        bool       mpress;

        void         editThisField();
     public:
        WShipArgWidget(int id, const GMission& mission, GMission::Parameter which, int sid);
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);
        bool handleEvent(const UIEvent& e, bool second_pass);
        void onStateChange(int flags, bool enable);
    };
}

/** Create a widget.
    \param id      widget Id
    \param mission we're editing this mission's parameter....
    \param which   ...this one (0 or 1)...
    \param sid     ...on this ship (may be null). */
WShipArgWidget::WShipArgWidget(int id, const GMission& mission, GMission::Parameter which, int sid)
    : UINumberSelector(id, 0, getArgLimit(mission.getParameterType(which)), 10),
      mission(mission), which(which), sid(sid),
      mpress(false)
{ }

void
WShipArgWidget::drawContent(GfxCanvas& can)
{
    string_t value;
    int num = getValue();
    if (!num) {
        value = _("<not set>");
    } else {
        GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        switch (mission.getParameterType(which) & GMission::ma_TypeMask) {
         case GMission::mat_PID:
         case GMission::mat_BID:
            if (univ.ty_any_planets.isValidIndex(num))
                value = univ.ty_any_planets.getObjectByIndex(num).getName(GObject::PlainName);
            break;
         case GMission::mat_SID_here:
         case GMission::mat_SID:
            if (univ.ty_any_ships.isValidIndex(num))
                value = univ.ty_any_ships.getObjectByIndex(num).getName(GObject::PlainName);
            break;
         case GMission::mat_Player:
            if (num > 0 && num <= NUM_PLAYERS)
                value = player_racenames.getShortName(num);
            break;
        }
        if (value.empty())
            value = numToString(num);
    }

    /* display it */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setSolidBackground();
    ctx.useFont(*fonts[FONT_TITLE]);
    if (getFocusState() != NoFocus) {
        ctx.setColor(UIColor::tc_Blue);
        outTextF(ctx, getExtent().x, getExtent().y, getExtent().w, value + "_");
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, getExtent().x, getExtent().y, getExtent().w, value);
    }
    can.update(getExtent());
}

void
WShipArgWidget::onStateChange(int state, bool enable)
{
    if (state & st_Focused)
        drawWidget();
    if (!enable && (state & st_Selected))
        mpress = 0;
    UINumberSelector::onStateChange(state, enable);
}

void
WShipArgWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_FlexibleH;
    info.min_size = font_sizes[FONT_TITLE].scaledBy(5, 1);
    info.pref_size = font_sizes[FONT_TITLE].scaledBy(15, 1);
}

/** Edit this field. */
void
WShipArgWidget::editThisField()
{
    int value = getValue();
    if (getMissionArg(mission, value, which, sid))
        setValue(value);
}

bool
WShipArgWidget::handleEvent(const UIEvent& e, bool /*second_pass*/)
{
    if (e.type == ev_Mouse && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        setSelectedWidget();
        if (e.mouse.b) {
            setFocusedWidget();
            mpress = true;
        } else if (mpress) {
            mpress = false;
            editThisField();
        }
        return true;
    } else if (e.type == ev_Key && hasState(st_Focused)) {
        if (e.key.code == ' ' || e.key.code == SDLK_TAB) {
            editThisField();
            return true;
        }
    }
    return false;     // (!) this is a hack, but PCC 1.x has this too :)
}

static UINumberSelector*
createWidget(const GMission& mission, TMissionData& output, GMission::Parameter which, int sid)
{
    UINumberSelector* p;
    if ((mission.getParameterType(which) & GMission::ma_TypeMask) == GMission::mat_Integer) {
        p = new UIDecimalSelector(1000+which, 0, 10000, 10);
    } else {
        p = new WShipArgWidget(1000+which, mission, which, sid);
    }
    p->setValue(output.args[which]);
    return p;
}

/** Get arguments for a two-arg mission.
    \param mission    the mission we're setting
    \param output     output will be placed here. Should be pre-initialized
                      correctly.
    \param sid        the ship on which we're setting the mission. May be 0. */
static bool
getTwoMissionArgs(const GMission& mission, TMissionData& output, int sid)
{
    ObjectHolder h;
    UIWindow win(mission.getName(), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);

    UINumberSelector& i = h.add(createWidget(mission, output, GMission::InterceptParameter, sid));
    UINumberSelector& t = h.add(createWidget(mission, output, GMission::TowParameter, sid));

    UIGroup& g = h.add(new UIGroup(h.add(new UIGridLayout(2, 5))));
    g.add(h.add(new UIStaticText(mission.getParameterName(GMission::InterceptParameter) + ":", UIStaticText::StaticTitle)));
    g.add(i);
    g.add(h.add(new UIStaticText(mission.getParameterName(GMission::TowParameter) + ":", UIStaticText::StaticTitle)));
    g.add(t);
    win.add(g);
    win.add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Vert)));
    if ((mission.getParameterType(GMission::InterceptParameter) & GMission::ma_TypeMask) != GMission::mat_Integer
        || (mission.getParameterType(GMission::TowParameter) & GMission::ma_TypeMask) != GMission::mat_Integer)
        win.add(h.add(new UIStaticText(_("Press space on a field to change it."),
                                       UIStaticText::StaticNormal)));

    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g2.add(h.add(new UIButton(1, _("OK"),     SDLK_RETURN)).addStop());
    g2.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g2.add(h.add(new UISpacer()));
    win.add(g2);
    win.pack();

    ui_root->centerWidget(win);
    ui_root->add(win);
    bool rv = ui_root->run();
    ui_root->close(win);
    if (rv) {
        output.args[0] = i.getValue();
        output.args[1] = t.getValue();
        return true;
    }
    return false;
}

/** Choose mission for a ship.
    \param list     list of all applicable missions
    \param sid      the ship on which we're setting a mission. May be null.
    \param output   output is placed here.
    \return true if user chose OK */
bool
chooseMission(const GMissionList& list, int sid, TMissionData& output)
{
    enum { EXT_MISSION_MARKER = -99 };

    UIStringList str_list;
    for (GMissionList::iterator i = list.begin(); i != list.end(); ++i)
        str_list.add(i->getNumber(), i->getHotkey() + string_t(" - ") + i->getName());
    str_list.add(EXT_MISSION_MARKER, _("# - Extended Mission (M.I.T.)"));

    UIStandardListbox box(0, 12, 340, str_list);
    if (sid)
        box.scrollToKey(getDisplayedTurn().getCurrentUniverse().getShip(sid).getMission());
    if (!box.doStandardListbox(_("Set Mission"), /* FIXME: */ 0))
        return false;

    int new_mission = box.getKey();

    if (sid != 0
        && new_mission == GMission::msn_Tow
        && getDisplayedTurn().getCurrentUniverse().isShipTowed(sid))
    {
        if (!UIMessageBox(_("This ship is already being towed. Tow chains will not work.\n"
                            "Continue anyway?"),
                          _("Set Mission")).doYesNoDialog())
        {
            return false;
        }
    }

    if (new_mission == EXT_MISSION_MARKER) {
        WExtendedMission window(output, sid);
        ui_root->centerWidget(window);
        ui_root->add(window);
        bool rv = ui_root->run();
        ui_root->close(window);
        return rv;
    } else {
        const GMission* msn = list.getMissionByNumber(new_mission);
        ASSERT(msn);

        /* Pre-initialize output */
        output.mission = new_mission;
        output.args[0] = output.args[1] = 0;
        if (sid && output.mission == getDisplayedTurn().getCurrentUniverse().getShip(sid).getMission()) {
            GShip& s = getDisplayedTurn().getCurrentUniverse().getShip(sid);
            output.args[0] = s.getInterceptId();
            output.args[1] = s.getTowId();
        }

        if (msn->getParameterType(GMission::InterceptParameter) && msn->getParameterType(GMission::TowParameter)) {
            // mission with two args
            return getTwoMissionArgs(*msn, output, sid);
        } else if (msn->getParameterType(GMission::InterceptParameter)) {
            // mission with Intercept arg
            return getMissionArg(*msn, output.args[0], GMission::InterceptParameter, sid);
        } else if (msn->getParameterType(GMission::TowParameter)) {
            // mission with Tow arg
            return getMissionArg(*msn, output.args[1], GMission::TowParameter, sid);
        } else {
            // mission without args
            return true;
        }
    }
}

/** Main entry: Edit mission on a ship. */
void
doShipMissionChange(GShip& ship)
{
    GMissionList msn_list (GMissionList::getInstance(), ship);
    TMissionData data;
    data.mission = ship.getMission();
    data.args[0] = ship.getInterceptId();
    data.args[1] = ship.getTowId();
    if (!chooseMission(msn_list, ship.getId(), data))
        return;
    // FIXME: fleet error handling
    setFleetMission(getDisplayedTurn().getCurrentUniverse(), ship.getId(), data.mission, data.args[0], data.args[1]);
}
