/**
  *  \file client/dlg-visual.cc
  *  \brief Visual Scanner Dialog
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chart.h"
#include "client/dlg-visual.h"
#include "client/scr-sim.h"
#include "cpluslib/format.h"
#include "cpluslib/sigslot.h"
#include "game/ship.h"
#include "game/sim-game.h"
#include "game/sim.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "resmgr/resid.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/util.h"
#include "ui/window.h"
#include "util/nls.h"
#include "game/univ.h"
#include "game/planet.h"
#include "client/actions.h"
#include "game/team.h"
#include "game/racename.h"
#include "game/config.h"
#include "client/remote.h"
#include "client/widgets/specview.h"
#include "client/tiles/visheader.h"
#include "client/tiles/vishull.h"
#include "client/tiles/visship.h"
#include "ui/widgets/imagewidget.h"
#include "game/global.h"
#include "resmgr/resmgr.h"
#include "client/wcolorframe.h"
#include "ui/widgets/button.h"
#include "client/dialogs/helpdlg.h"
#include "client/widgets/objlist.h"
#include "game/objsel.h"

namespace {
    enum {
        cm_VisSelect,
        cm_VisCancel,
        cm_VisAdd
    };

    class WVisualScanWindow : public UIWindow, private GObjectSelectionObserver {
        GObjectList& list;
        GObjectSelection& selection;
        SignalConnection conn;

        ObjectHolder h;

        void init(const string_t& okname);

        // GObjectSelectionObserver:
        void onObjectChanged();

     public:
        enum Mode {
            NormalMode,
            SpecMode,
            ListMode
        };

        WVisualScanWindow(string_t title, string_t okname, GObjectList& list, GObjectSelection& selection);
        ~WVisualScanWindow();
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        bool handleEvent(const UIEvent& e, bool second_pass);
        GObjectReference getCurrentObjectReference() const;
        void setIndex(int n);
        void setMode(Mode mode);
        GObjectList& getList()
            { return list; }
        GObjectSelection& getSelection()
            { return selection; }
        void setAcceptForeignShips(bool v)
            { allow_foreign = v; }

     private:
        UIImageWidget image;   // FIXME?
        WColorFrame   image_cf;
        WColorFrame*  button_r_cf;
        UIBaseWidget* button_l;
        UIBaseWidget* button_s;
        UIBaseWidget* peer;
        Mode          mode;
        bool allow_foreign;
    };

    class WVisualScanSpecInfoWindow : public UIWindow, private GObjectSelectionObserver {
        ObjectHolder h;
        UIBaseWidget& peer;
        WSpecView* spec_view;
     public:
        WVisualScanSpecInfoWindow(GObjectSelection& selection, UIBaseWidget& peer);
        ~WVisualScanSpecInfoWindow();
        bool handleEvent(const UIEvent& event, bool second_pass);
        void init();
        void onObjectChanged();
    };

    class WVisualScanListWindow : public UIWindow {
        ObjectHolder h;
        WVisualScanWindow& window;
        WObjectList* list;
        SignalConnection conn_scroll_list;
        SignalConnection conn_scroll_window;
        SignalConnection conn_menu;
     public:
        WVisualScanListWindow(WVisualScanWindow& window);
        ~WVisualScanListWindow();
        bool handleEvent(const UIEvent& event, bool second_pass);
        void init();
        void onScrollList();
        void onScrollWindow();
        void onMenu();
    };
}

static WColorFrame&
createButton(UIGroup& container, ObjectHolder& holder, const char* title, int key)
{
    UIButton& btn   = holder.add(new UIButton(title, key));
    WColorFrame& cf = holder.add(new WColorFrame());
    cf.add(btn);
    container.add(cf);
    return cf;
}

/*************************** WVisualScanWindow ***************************/

WVisualScanWindow::WVisualScanWindow(string_t title, string_t okname, GObjectList& list, GObjectSelection& selection)
    : UIWindow(title, BLUE_WINDOW, COLOR_GRAY),
      GObjectSelectionObserver(selection),
      list(list),
      selection(selection),
      image(105+2, 93+2),
      button_r_cf(0),
      button_l(0),
      button_s(0),
      peer(0),
      mode(NormalMode),
      allow_foreign(false)
{
    init(okname);
}

void
WVisualScanWindow::init(const string_t& okname)
{
    // Widget tree:
    //    Window [VBox]
    //      Group [HBox]      1
    //        Group [VBox]    11
    //          HeaderTile
    //          Group [HBox]  111
    //            Picture
    //            HullInfoTile
    //          ShipInfoTile
    //        Group [VBox]    12
    //          Buttons "Prev", "S", "L", "Next"
    //      Group [HBox]      2
    //        Buttons "OK", "Ins", "H", etc.

    // Button 'r' needs a color frame. To align all buttons in line, all buttons
    // in group 12 receive a WColorFrame. Since the color frame is 2 pixels, we
    // use a VBox layout with offset 1 to achieve the visual appearance of the
    // usual offset 5.

    setLayout(UIVBoxLayout::instance5);

    UIGroup& group1   = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group11  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group111 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group12  = h.add(new UIGroup(h.add(new UIVBoxLayout(1))));
    UIGroup& group2   = h.add(new UIGroup(UIHBoxLayout::instance5));

    add(group1);
    add(group2);

    group1.add(group11);
    group1.add(group12);

    group11.add(h.add(new WVisualScanHeaderTile(selection)));
    group11.add(group111);
    group11.add(h.add(new WVisualScanShipInfoTile(selection)));

    image_cf.add(image);
    group111.add(image_cf);
    group111.add(h.add(new WVisualScanHullInfoTile(selection)));

    createButton(group12, h, "\xE2\x86\x91", '-');
    if (config.CPEnableRemote()) {
        button_r_cf = &createButton(group12, h, "R", 'r');
    }
    group12.add(h.add(new UISpacer()));
    button_l = createButton(group12, h, "L", 'l').getFirstChild();
    button_s = createButton(group12, h, "S", 's').getFirstChild();

    createButton(group12, h, "\xE2\x86\x93", '+');

    group2.add(h.add(new UIButton(cm_VisSelect, okname, SDLK_RETURN)).addCommand());
    group2.add(h.add(new UIButton(cm_VisCancel, _("ESC"), SDLK_ESCAPE)).addCommand());
    group2.add(h.add(new UIButton(cm_VisAdd, _("Add"), SDLK_INSERT)).addCommand());
    group2.add(h.add(new UISpacer()));
    group2.add(h.add(new UIButton(_("Help"), 'h')));

    onObjectChanged();
}

WVisualScanWindow::~WVisualScanWindow()
{
    delete peer;
}

/** Object changed (scrolled or modified). */
void
WVisualScanWindow::onObjectChanged()
{
    // Redraw WVisualScanWindow (color frames):
    drawWidget(false);

    // Update image:
    if (GShip* s = dynamic_cast<GShip*>(getCurrentObject())) {
        mp16_t h = s->getHullId();
        if (h.isKnown())
            image.setImage(requestPixmap(ResId(res::ship, getHull(h).getInternalPictureNumber(), h)));
        else
            image.setImage(requestPixmap(ResId("nvc")));
        if (button_r_cf != 0)
            button_r_cf->setColor(getRemoteControlColor(getNewRemoteControlState(*s, getDisplayedTurn().getCommands(getPlayerId()))));
    } else {
        image.setImage(0);
        if (button_r_cf != 0)
            button_r_cf->setColor(cf_None);
    }

    // Selection frame:
    image_cf.setColor(getCurrentObjectReference()->isMarked() ? cf_Yellow : cf_None);
}

bool
WVisualScanWindow::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case cm_VisSelect:
        if (allow_foreign || getCurrentObjectReference()->isPlayed())
            stop(1);
        return true;
     case cm_VisCancel:
        stop(0);
        return true;
     case cm_VisAdd:
        if (GShip* s = dynamic_cast<GShip*>(getCurrentObject()))
            doShipAddToSim(*s, true);
        else if (GPlanet* p = dynamic_cast<GPlanet*>(getCurrentObject()))
            doPlanetAddToSim(*p);
        return true;
     default:
        return UIWindow::handleCommand(cmd, sender);
    }
}

bool
WVisualScanWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    int n;
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case '+':
         case SDLK_DOWN:
         case SDLK_PAGEDOWN:
         case vk_WheelDown:
         case ss_Ctrl + '+':
         case ss_Ctrl + SDLK_DOWN:
         case ss_Ctrl + SDLK_PAGEDOWN:
         case ss_Ctrl + vk_WheelDown:
            setSelectedWidget();
            if ((n = list.findNextIndexWrap(selection.getCurrentIndex(), (e.key.code & ss_Ctrl) != 0)) > 0)
                setIndex(n);
            return true;

         case '-':
         case SDLK_UP:
         case SDLK_PAGEUP:
         case vk_WheelUp:
         case ss_Ctrl + '-':
         case ss_Ctrl + SDLK_UP:
         case ss_Ctrl + SDLK_PAGEUP:
         case ss_Ctrl + vk_WheelUp:
            setSelectedWidget();
            if ((n = list.findPreviousIndexWrap(selection.getCurrentIndex(), (e.key.code & ss_Ctrl) != 0)) > 0)
                setIndex(n);
            return true;

         case '.':
            if (getCurrentObjectReference().isGameObject()) {
                GObject& o = *getCurrentObject();
                o.setIsMarked(!o.isMarked());
                getCurrentUniverse()->doScreenUpdates();
            }
            return true;

         case 'r':
            if (GShip* sh = dynamic_cast<GShip*>(getCurrentObject())) {
                doRemoteControl(*sh, getDisplayedTurn().getCommands(getPlayerId()));
                getCurrentUniverse()->doScreenUpdates();
            }
            return true;

         case 's':
            setMode(mode == SpecMode ? NormalMode : SpecMode);
            return true;

         case 'l':
            setMode(mode == ListMode ? NormalMode : ListMode);
            return true;

         case 'h':
         case ss_Alt + 'h':
            doHelp("pcc2:listship");
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

GObjectReference
WVisualScanWindow::getCurrentObjectReference() const
{
    int index = selection.getCurrentIndex();
    if (index > 0 && index <= list.size())
        return list.getObjectReferenceByIndex(index);
    else
        return GObjectReference();
}

void
WVisualScanWindow::setIndex(int n)
{
    if (n > list.size())
        n = list.size();
    if (n <= 0)
        n = 1;
    if (n != selection.getCurrentIndex()) {
        selection.setCurrentIndex(n);
    }
}

void
WVisualScanWindow::setMode(Mode mode)
{
    if (mode != this->mode) {
        this->mode = mode;

        // Delete old peer
        if (peer) {
            delete peer;
            peer = 0;
        }

        // Make new peer
        if (mode == SpecMode) {
            peer = new WVisualScanSpecInfoWindow(selection, *this);
            ui_root->add(*peer);
        }
        if (mode == ListMode) {
            peer = new WVisualScanListWindow(*this);
            ui_root->add(*peer);
        }

        if (button_s) {
            button_s->setState(UIButton::bf_Highlight, mode == SpecMode);
        }
        if (button_l) {
            button_l->setState(UIButton::bf_Highlight, mode == ListMode);
        }
    }
}

/*********************** WVisualScanSpecInfoWindow ***********************/

WVisualScanSpecInfoWindow::WVisualScanSpecInfoWindow(GObjectSelection& selection, UIBaseWidget& peer)
    : UIWindow(_("Ship Specification"), BLUE_WINDOW, COLOR_DARK),
      GObjectSelectionObserver(selection),
      peer(peer),
      spec_view(0)
{
    init();
}

WVisualScanSpecInfoWindow::~WVisualScanSpecInfoWindow()
{ }

bool
WVisualScanSpecInfoWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    return UIWindow::handleEvent(event, second_pass)
        || peer.handleEvent(event, second_pass);
}

void
WVisualScanSpecInfoWindow::init()
{
    /* Generate content */
    setLayout(UIVBoxLayout::instance5);
    spec_view = &h.add(new WSpecView());
    add(*spec_view);

    /* Do layout */
    pack();
    setOrigin(peer.getExtent().getRightX(), peer.getExtent().y);

    /* Perform first update */
    onObjectChanged();
}

void
WVisualScanSpecInfoWindow::onObjectChanged()
{
    GShipQuery query;
    if (GShip* s = dynamic_cast<GShip*>(getCurrentObject())) {
        query.initForExistingShip(*getCurrentUniverse(), s->getId());
    } else {
        query.clear();
    }
    spec_view->setQuery(query);
}

/************************* WVisualScanListWindow *************************/

WVisualScanListWindow::WVisualScanListWindow(WVisualScanWindow& window)
    : UIWindow(_("Ship List"), BLUE_WINDOW, COLOR_GRAY),
      window(window),
      list(0)
{
    init();
}

WVisualScanListWindow::~WVisualScanListWindow()
{ }

bool
WVisualScanListWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && (event.key.code == '#' || event.key.code == '\\')) {
        onMenu();
        return true;
    }
    return UIWindow::handleEvent(event, second_pass)
        || window.handleEvent(event, second_pass);
}

void
WVisualScanListWindow::init()
{
    /* Generate content */
    setLayout(UIHBoxLayout::instance0);
    list = &h.add(new WObjectList(1, 25 /* FIXME */, 20 * font_ems[FONT_NORMAL], window.getList()));
    add(*list);
    add(h.add(new UIScrollbar(*list)));

    /* Do layout */
    pack();
    setOrigin(window.getExtent().getRightX(), window.getExtent().y);

    /* Register ourselves for updates */
    conn_scroll_window = window.getSelection().sig_current_changed.add(this, &WVisualScanListWindow::onScrollWindow);
    conn_scroll_list   = list->sig_moved.add(this, &WVisualScanListWindow::onScrollList);
    conn_menu          = list->sig_menu.add(this, &WVisualScanListWindow::onMenu);

    onScrollWindow();
}

void
WVisualScanListWindow::onScrollList()
{
    window.setIndex(list->getCurrent()+1);
}

void
WVisualScanListWindow::onScrollWindow()
{
    list->selectObject(window.getCurrentObjectReference());
}

void
WVisualScanListWindow::onMenu()
{
    // Build menu
    StringList m;
    m.add(SortById,       _("Sort by Id"));
    m.add(SortByName,     _("Sort by Name"));
    m.add(SortByOwner,    _("Sort by Owner"));
    m.add(SortByHull,     _("Sort by Hull Type"));
    m.add(SortByHullMass, _("Sort by Hull Mass"));
    m.add(SortByMass,     _("Sort by Mass"));
    m.add(SortByDamage,   _("Sort by Damage"));
    m.add(SortByFleet,    _("Sort by Fleet"));
    m.add(SortByTowGroup, _("Sort by Tow Group"));

    UIStandardListbox mbox(0, m.getCount(), getTextWidth(*fonts[FONT_NORMAL], m) + 20, m);
    mbox.scrollToKey(getSortOrderConfig());
    if (mbox.doStandardMenu(*list)) {
        GObjectReference id = list->getCurrentObject();
        setSortOrderConfig(mbox.getKey());
        window.getList().postprocessUser(true);
        list->handleListChange(id);
    }
}

/** Do standard "list ships" dialog.
    \param pt        location to list
    \param flags     flags (ls_XXX)
    \param exclude_ship do not show that ship
    \param title     title line
    \param okname    name of "ok" button */
GObjectReference
doListShips(GPoint pt, int flags, int exclude_ship,
            string_t title, string_t okname)
{
    /* Initialize list */
    GObjectList l;
    l.addObjectsAt(getDisplayedTurn().getCurrentUniverse(), pt, flags, exclude_ship);
    l.postprocessUser(true);
    int n = l.countGameObjects();

    /* List empty? Show message. */
    if (n == 0) {
        string_t msg;
        const GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        if (exclude_ship != 0
            && (!univ.ty_any_ships.isValidIndex(exclude_ship)
                || univ.getShip(exclude_ship).getPos() != pt))
        {
            /* we were told to exclude a ship, but that ship isn't here */
            exclude_ship = 0;
        }

        if (!(flags & ls_IncludeForeignShips)) {
            if (exclude_ship)
                msg = _("There is no other ship of ours at that position.");
            else
                msg = _("There is no ship of ours at that position.");
        } else {
            if (exclude_ship) {
                msg = _("We can't locate another ship at that position.");
            } else {
                msg = _("We can't locate a ship at that position.");
                int pid = univ.getPlanetAt(pt);
                if (pid && !univ.getPlanet(pid).isPlayed())
                    /* this message must start with a space */
                    msg += format(_(" The planet %s may be hiding ships from our sensors."),
                                  univ.getPlanet(pid).getName(GObject::PlainName));
            }
        }
        messageBox(msg, _("Scanner"));
        return GObjectReference();
    }

    /* One object only? Bail out early if allowed. */
    if (n == 1 && (flags & ls_EarlyExit)) {
        GObjectList::size_type n = 1;
        while (!l.getObjectReferenceByIndex(n).isGameObject())
            ++n;
        if ((flags & ls_AllowForeignShips) || l.getObjectReferenceByIndex(n)->isPlayed())
            return l.getObjectReferenceByIndex(n);
    }

    /* Create an object selection to manage the current position */
    GSimpleObjectSelection selection;
    selection.setObjectType(&l);

    /* Do dialog. */
    WVisualScanWindow window(title, okname, l, selection);
    window.setAcceptForeignShips((flags & ls_AllowForeignShips) != 0);
    window.pack();
    ui_root->moveWidgetToEdge(window, 0, 0, 5);
    int result = ui_root->runModal(window);

    // window.setMode(window.NormalMode);

    if (result)
        return window.getCurrentObjectReference();
    else
        return GObjectReference();
}
