/**
  *  \file game/planet.cc
  *  \brief Planet / Starbase Management
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/planet.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/univ.h"
#include "game/planet-int.h"
#include "game/ship.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/host.h"
#include "game/team.h"
#include "util/recode.h"
#include "int/data.h"
#include "int/intglobal.h"

static const TPlanet blank_planet = {
    UNKNOWNP,                               // owner
    UNKNOWNP,                               // planet_id
    { UNKNOWNC, UNKNOWNC, UNKNOWNC },       // fc
    UNKNOWNP, UNKNOWNP, UNKNOWNP,           // mi, fa, de
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ore mined
    UNKNOWNP,                               // colonists
    UNKNOWNP, UNKNOWNP,                     // sup, mc
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ground ore
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ore density
    UNKNOWNP, UNKNOWNP,                     // ctax, ntax
    UNKNOWNS, UNKNOWNS,                     // chap, nhap
    UNKNOWNP, UNKNOWNP, UNKNOWNP,           // ngov, natives, nrace
    UNKNOWNP,                               // temp
    UNKNOWNP                                // base
};

static const TStarbase blank_starbase = {
    UNKNOWNP,                                    // base_id
    UNKNOWNP,                                    // owner
    UNKNOWNP,                                    // defense
    UNKNOWNP,                                    // damage
    { UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP },  // tech
    { UNKNOWNP, UNKNOWNP, UNKNOWNP,              // engine_store
      UNKNOWNP, UNKNOWNP, UNKNOWNP,
      UNKNOWNP, UNKNOWNP, UNKNOWNP },
    { UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP,    // hull_store
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP,
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP,
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP,
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP },
    { UNKNOWNP, UNKNOWNP, UNKNOWNP,              // beam_store
      UNKNOWNP, UNKNOWNP, UNKNOWNP,
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP },
    { UNKNOWNP, UNKNOWNP, UNKNOWNP,              // launcher_store
      UNKNOWNP, UNKNOWNP, UNKNOWNP,
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP },
    { UNKNOWNP, UNKNOWNP, UNKNOWNP,              // ammo_store
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP,
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP },
    UNKNOWNP, UNKNOWNP,                          // shipyard_id, shipyard_action
    UNKNOWNP,                                    // mission
    { UNKNOWNP, UNKNOWNP, UNKNOWNP,              // build
      UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP }
};

static const TAutobuildSettings standard_autobuild = {
    { 1000, 1000, 1000, 20 },
    { 5, 10, 3, 2 }
};

/************************** Map/Planet/BaseInfo **************************/

GPlanet::MapInfo::MapInfo(int id)
    : id(id),
      name("?"),
      position(0, 0),
      known_to_not_exist(false),
      base_kind(UnknownBase),
      planet_kind(NoPlanet),
      orbit_flags(0),
      autobuild(standard_autobuild)
{ }

GPlanet::MapInfo::~MapInfo()
{ }

GPlanet::PlanetInfo::PlanetInfo(int id)
    : data(blank_planet),
      known_to_have_natives(false),
      industry_level()
{
    data.planet_id = id;
    time[ts_Minerals] = time[ts_Colonists] = time[ts_Natives] = time[ts_Cash] = 0;
}

GPlanet::BaseInfo::BaseInfo(int id)
    : data(blank_starbase),
      queue_pos(),
      queue_priority()
{
    data.base_id = id;
}

/******************************** GPlanet ********************************/

/** Constructor. Create blank planet of given Id.
    \param id Planet Id */
GPlanet::GPlanet(int id)
    : GMapObject(),
      unit_scores(),
      cargo_arbiter(),
      planet_info(0),
      base_info(0),
      map_info(id),
      planet_source(),
      base_source()
{ }

/** Copy constructor. */
GPlanet::GPlanet(const GPlanet& p)
    : GMapObject(),
      unit_scores(p.unit_scores),
      cargo_arbiter(),
      planet_info(0),
      base_info(0),
      map_info(p.map_info),
      planet_source(p.planet_source),
      base_source(p.base_source)
{
    if (p.planet_info) {
        planet_info = new PlanetInfo(*p.planet_info);
    }
    if (p.base_info) {
        base_info = new BaseInfo(*p.base_info);
    }
    setIsMarked(p.isMarked());
    setPlayability(p.getPlayability());
}

/** Destructor. */
GPlanet::~GPlanet()
{
    delete planet_info;
    delete base_info;
}


/*
 *  Load and Save
 */

/** Add .dat file entry.
    This assumes that when we see a planet through several .dat files, we get the same one each time.
    \param dat parsed .dat file entry
    \param source source flag value to use for this entry */
void
GPlanet::addPlanetData(const TPlanet& dat, GPlayerSet source)
{
    // FIXME: older PHost versions clear the following fields of a
    // planet when sending a pdata planet target for exploration of an
    // unowned planet (owned planets never generate a target):
    // colonists, supplies, credits, mines, factories, defense,
    // coltax, nattax. We may want to merge that information somehow,
    // or detect and/or merge it.
    createPlanetInfo().data = dat;
    planet_source |= source;
}

/** Add starbase .dat file entry. See addPlanetData() for details. */
void
GPlanet::addBaseData(const TStarbase& bdat, GPlayerSet source)
{
    if (!base_info) {
        base_info = new BaseInfo(map_info.id);
    }
    base_info->data = bdat;
    base_source |= source;
}

/** Add history file entry. */
void
GPlanet::addHistoryData(const TDbPlanet& data)
{
    if (planet_source.empty()) {
        // Accept only when not scanned otherwise
        PlanetInfo& pi = createPlanetInfo();

        // FIXME: This assumes only one history data source. Otherwise, we have to merge.
        pi.data               = data.planet;
        pi.time[ts_Minerals]  = data.time[ts_Minerals];
        pi.time[ts_Colonists] = data.time[ts_Colonists];
        pi.time[ts_Natives]   = data.time[ts_Natives];
        pi.time[ts_Cash]      = data.time[ts_Cash];
        if (data.planet.factories >= 30000) {
            pi.data.factories = mp16_t().getRawValue();
            pi.industry_level = mp16_t(data.planet.factories - 30000);
        } else {
            pi.industry_level = mp16_t();
        }
        pi.known_to_have_natives = (data.known_to_have_natives != 0);
    }
}

/** Set planet position. */
void
GPlanet::setXY(int x, int y)
{
    map_info.position.x = x;
    map_info.position.y = y;
}

/** Set planet name. */
void
GPlanet::setName(const string_t& name)
{
    map_info.name = name;
}

/** Set whether non-existance of this planet is known. There is no way
    to explicitly specify that a planet does not exist. To build maps
    with fewer than 500 planets, people move planets to odd positions.
    Recent PHosts send a util.dat message whenever they consider a
    planet to be non-existant, to make sure that the clients' idea of
    which planets do exist agrees with PHost's. */
void
GPlanet::setKnownToNotExist(bool value)
{
    map_info.known_to_not_exist = value;
}

/** Add autobuild record from database file.
    \param abs [in] Autobuild record */
void
GPlanet::setAutobuild(const TAutobuildSettings& abs)
{
    map_info.autobuild = abs;
}

/** Get planet data record for storage in data files.
    \param dat [out] planet data */
void
GPlanet::getPlanetData(TPlanet& dat) const
{
    dat = getDisplayedPlanet();
}

/** Get starbase data record for storage in data files.
    \param bdat [out] base data */
void
GPlanet::getBaseData(TStarbase& bdat) const
{
    bdat = getDisplayedBase();
}

/** Get Autobuild record for storage in database file.
    \param abs [out] Autobuild record */
void
GPlanet::getAutobuild(TAutobuildSettings& abs) const
{
    abs = map_info.autobuild;
}

/** Get planet history record for storage in database file.
    \param data [out] History record */
void
GPlanet::getHistoryData(TDbPlanet& data) const
{
    if (planet_info != 0) {
        data.planet             = planet_info->data;
        data.time[ts_Minerals]  = planet_info->time[ts_Minerals];
        data.time[ts_Colonists] = planet_info->time[ts_Colonists];
        data.time[ts_Natives]   = planet_info->time[ts_Natives];
        data.time[ts_Cash]      = planet_info->time[ts_Cash];
        data.known_to_have_natives = planet_info->known_to_have_natives;
        if (planet_info->industry_level.isKnown() && mp16_t(planet_info->data.factories).isKnown()) {
            data.planet.factories = planet_info->industry_level + 30000;
        }
        if (base_info) {
            data.planet.build_base = 1;
        }
    } else {
        data.planet                = blank_planet;
        data.time[ts_Minerals]     = data.time[ts_Colonists] = data.time[ts_Natives] = data.time[ts_Cash] = 0;
        data.known_to_have_natives = 0;
    }
}

/** Do internal checks for this planet.
    Internal checks do not require a partner to interact with.
    This will fix the problems, and display appropriate messages. */
void
GPlanet::internalCheck()
{
    map_info.orbit_flags = 0;

    // Does this planet exist?
    bool exists = !map_info.known_to_not_exist && chart_conf.isValidPlanetCoordinate(map_info.position);

    // Check PDATA. If we have a PDATA entry, it must exist.
    if (planet_source.nonempty() && !exists) {
        const char* fmt = (map_info.known_to_not_exist
                           ? _("Planet #%d has data although it is reported as non-existant -- host confused?")
                           : _("Planet #%d exists for Host, but is outside valid range"));
        console.write(LOG_WARN, format(fmt, map_info.id));
        exists = true;
    }

    // Set planet kind
    if (!exists) {
        if (planet_info != 0) {
            map_info.planet_kind = HiddenPlanet;
        } else {
            map_info.planet_kind = NoPlanet;
        }
    } else if (planet_source.nonempty()) {
        map_info.planet_kind = CurrentPlanet;
    } else if (planet_info != 0) {
        map_info.planet_kind = KnownPlanet;
    } else {
        map_info.planet_kind = UnknownPlanet;
    }

    // Check BDATA. If we have BDATA, we also must have PDATA.
    if (base_source.nonempty() && planet_source.empty()) {
        // FIXME this will make PCC2 write invalid files, i.e. where BDATA.DAT and BDATA.DIS disagree.
        console.write(LOG_WARN, format(_("Starbase #%d does not have a planet -- deleting it"), map_info.id));
        delete base_info;
        base_info = 0;
        base_source = GPlayerSet();
    }

    // Set base kind
    if (base_source.nonempty()) {
        map_info.base_kind = CurrentBase;
    } else if (base_info != 0) {
        map_info.base_kind = KnownBase;
    } else if (planet_info != 0 && planet_info->data.build_base > 0) {
        map_info.base_kind = ExistingBase;
    } else if (planet_info != 0 && planet_info->data.build_base == 0) {
        map_info.base_kind = NoBase;
    } else {
        map_info.base_kind = UnknownBase;
    }
}

/** Combined checks, phase 2.
    This will do all post-processing which needs a partner to interact with.
    It requires the playability to be filled in. */
void
GPlanet::combinedCheck2(GUniverse& univ)
{
    if (planet_info != 0) {
        if (map_info.planet_kind == CurrentPlanet) {
            // We have seen this planet this turn
            planet_info->time[ts_Minerals]
                = planet_info->time[ts_Colonists]
                = planet_info->time[ts_Natives]
                = planet_info->time[ts_Cash]
                = univ.getTurnNumber();
        } else {
            // We have taken this planet from the history
            if (isOwnerKnown() && univ.isData(getOwner())) {
                /* planet is played by us, but we do no longer own it */
                planet_info->data.owner = 0;
            }
        }

        if (isOwnerKnown() && univ.isData(getOwner()) && map_info.base_kind != NoBase && map_info.base_kind != CurrentBase) {
            // We play this planet, and have history information about a base,
            // but that base isn't there.
            // -or- We're building a base here (this will set the status to ExistingBase).
            // Delete the base.
            map_info.base_kind = NoBase;
        }
    }
}

/** Get orbit flags. This classifies the ships in orbit, see FlagXXX. */
int
GPlanet::getOrbitFlags() const
{
    return map_info.orbit_flags;
}

/** Clear orbit flags. This will remove all planet ship marks. */
void
GPlanet::clearOrbitFlags()
{
    map_info.orbit_flags = 0;
}

/** Notify planet of ship in orbit. This will compute the planet markings. */
void
GPlanet::setShipInOrbit(const GShip& s)
{
    switch (getPlayerRelation(s.getOwner())) {
     case is_Me:
        map_info.orbit_flags |= FlagOwnShipsInOrbit;
        break;
     case is_Enemy:
        if (s.isReliablyVisible(0))
            map_info.orbit_flags |= FlagEnemyShipsInOrbit;
        else
            map_info.orbit_flags |= FlagGuessedEnemyInOrbit;
        break;
     case is_Ally:
        if (s.isReliablyVisible(0))
            map_info.orbit_flags |= FlagAlliedShipsInOrbit;
        else
            map_info.orbit_flags |= FlagGuessedAllyInOrbit;
        break;
    }
}


/*
 *  GMapObject interface
 */

string_t
GPlanet::getName(Name which) const
{
    switch (which) {
     case PlainName:
        return map_info.name;
     case LongName:
     case DetailedName:
        {
            string_t result = format(_("Planet #%d: %s"), map_info.id, map_info.name);
            if (which == DetailedName) {
                IntValue* iv = getPlanetProperties(map_info.id)[pp_Comment];
                if (getBoolValue(iv) > 0) {
                    result += ": ";
                    result += iv->toString(false);
                }
            }
            return result;
        }
    }
    return string_t();
}

int
GPlanet::getId() const
{
    return map_info.id;
}

/** Get owner. Note that this returns 0 for unknown owners. */
int
GPlanet::getOwner() const
{
    if (planet_info && mp16_t(planet_info->data.owner).isKnown()) {
        return planet_info->data.owner;
    } else {
        return 0;
    }
}

GPoint
GPlanet::getPos() const
{
    return map_info.position;
}


/*
 *  Planet Status Accessors:
 */

/** Check whether planet is visible. */
bool
GPlanet::isVisible() const
{
    return map_info.planet_kind != NoPlanet
        && map_info.planet_kind != HiddenPlanet;
}

/** Get planet source flags. This is the set of players whose PDATA
    file contains a copy of this planet (usually a unit set, but may
    be larger for unowned planets). */
GPlayerSet
GPlanet::getPlanetSource() const
{
    return planet_source;
}

/** Get planet kind. This determines that available information on the planet. */
GPlanet::PlanetKind
GPlanet::getPlanetKind() const
{
    return map_info.planet_kind;
}

/** Check whether we have any information about this planet.
    Note that the planet may not exist even if it has information (HiddenPlanet).
    \return true iff we have any information, full or partial */
bool
GPlanet::hasAnyPlanetData() const
{
    return planet_info != 0;
}

/** Check whether we have full planet data. */
bool
GPlanet::hasFullPlanetData() const
{
    return planet_source.nonempty();
}

/** Get history timestamp. */
int16_t
GPlanet::getHistoryTimestamp(TDbPlanetTimestamp which_one) const
{
    if (planet_info != 0) {
        return planet_info->time[which_one];
    } else {
        return 0;
    }
}


/*
 *  Base Status Accessors:
 */

/** Get base source flags. This is the set of players whose BDATA file
    contains a copy of this base (usually a unit set). */
GPlayerSet
GPlanet::getBaseSource() const
{
    return base_source;
}

/** Get starbase kind. This determines that available information on the starbase. */
GPlanet::BaseKind
GPlanet::getBaseKind() const
{
    return map_info.base_kind;
}

/** Check for starbase.
    \retval true this planet has a starbase
    \retval false this planet has no starbase, or we don't know */
bool
GPlanet::hasBase() const
{
    return map_info.base_kind != UnknownBase
        && map_info.base_kind != NoBase;
}

/** True iff we have full starbase information.
    \return true iff we have full, playable data. If yes, all base accessors will work. */
bool
GPlanet::hasFullBaseData() const
{
    return base_source.nonempty();
}



/*
 *  Owner accessors:
 *  - getOwner: returns 0 for unknown
 *  - setOwner: allows full mp16_t
 *  - isOwnerKnown: check for unknownness
 *
 *  Caveat: setOwner(getOwner()) may set the owner to "known as 0"
 */

/** Set owner.
    \param owner new owner */
void
GPlanet::setOwner(mp16_t owner)
{
    createPlanetInfo().data.owner = owner.getRawValue();
}

/** Check whether owner is known. */
bool
GPlanet::isOwnerKnown() const
{
    return mp16_t(getDisplayedPlanet().owner).isKnown();
}



/*
 *  Structure accessors
 */

/** Get number of structures.
    \param ps Which structure to query
    \return number of structures, or unknown */
mp16_t
GPlanet::getStructures(TPlanetaryStructure ps) const
{
    if (planet_info != 0) {
        switch (ps) {
         case str_Mines:
            return mp16_t(planet_info->data.mines);
         case str_Factories:
            return mp16_t(planet_info->data.factories);
         case str_Defense:
            return mp16_t(planet_info->data.defense);
         case str_BaseDefense:
            // FIXME: if we do not know whether there is a base, the right
            // answer would be mp16_t().
            return base_info ? mp16_t(base_info->data.defense) : mp16_t(0);
        }
    }
    return mp16_t();
}

/** Set number of structures. Works for own and scanned foreign planets.
    This cannot set base defense if the planet doesn't have a base.

    \param ps    Which structure to set
    \param count New amount */
void
GPlanet::setStructures(TPlanetaryStructure ps, mp16_t count)
{
    if (count != getStructures(ps)) {
        switch (ps) {
         case str_Mines:
            createPlanetInfo().data.mines = count.getRawValue();
            break;
         case str_Factories:
            createPlanetInfo().data.factories = count.getRawValue();
            break;
         case str_Defense:
            createPlanetInfo().data.defense = count.getRawValue();
            break;
         case str_BaseDefense:
            createBaseInfo().data.defense = count.getRawValue();
            break;
        }
        markDirty();
    }
}

/** Get industry level of this planet. Reports the industry level from
    known structure counts if available, otherwise from sensor scans. */
mp16_t
GPlanet::getIndustryLevel() const
{
    mp16_t mi = getStructures(str_Mines);
    mp16_t fa = getStructures(str_Factories);
    if (mi.isKnown() && fa.isKnown()) {
        return mp16_t(getIndustryLevel(mi + fa));
    } else if (planet_info) {
        return planet_info->industry_level;
    } else {
        return mp16_t();
    }
}

/** Get industry level for a given structure count.
    \param mifa Mines+Factories
    \return level (ind_XXX) */
int
GPlanet::getIndustryLevel(int mifa)
{
    if (host.isPHost()) {
        int level = mifa / 30;
        if (level > ind_Heavy) {
            return ind_Heavy;
        } else {
            return level;
        }
    } else {
        if (mifa >= 100) {
            return ind_Heavy;
        } else {
            return ind_Light;
        }
    }
}

/** Set industry level for this planet. This routine only makes sense
    for planets we do not play. */
void
GPlanet::setIndustryLevel(mp16_t level)
{
    // If we're playing this planet, setting the industry level is useless
    if (hasFullPlanetData())
        return;

    // We're not playing the planet, so set the level and check for
    // conflicts with our other data
    createPlanetInfo().industry_level = level;
    if (level.isKnown()) {
        mp16_t mi = getStructures(str_Mines);
        mp16_t fa = getStructures(str_Factories);
        if (mi.isKnown() && fa.isKnown()) {
            // We know mines and factories. Does this conflict with our level?
            if (getIndustryLevel(mi + fa) != level) {
                // yes, so forget mi+fa
                setStructures(str_Mines, mp16_t());
                setStructures(str_Factories, mp16_t());
            }
        } else if (mi.isKnown() || fa.isKnown()) {
            // We know mines or factories, but not both. Does this conflict with our level?
            int mifa = 0;
            if (mi.isKnown())
                mifa += mi;
            if (fa.isKnown())
                mifa += fa;
            if (getIndustryLevel(mifa) > level) {
                // our stored mine/factory count would yield a larger
                // level than reported. This means our stored mi/fa is
                // outdated.
                setStructures(str_Mines, mp16_t());
                setStructures(str_Factories, mp16_t());
            }
        }
    }
}


/*
 *  Colonist Accessors
 */

mn16_t
GPlanet::getColonistHappiness() const
{
    return mn16_t(getDisplayedPlanet().colonist_happy);
}

void
GPlanet::setColonistHappiness(mn16_t happiness)
{
    createPlanetInfo().data.colonist_happy = happiness.getRawValue();
    markDirty();
}

mp16_t
GPlanet::getColonistTax() const
{
    return mp16_t(getDisplayedPlanet().colonist_tax);
}

void
GPlanet::setColonistTax(mp16_t tax)
{
    createPlanetInfo().data.colonist_tax = tax.getRawValue();
    markDirty();
}

mn32_t
GPlanet::getColonists() const
{
    // This uses mn32_t because the reservation might exceed the amount.
    // If we do not know the amount, assume the reservation is empty. What else.
    uint32_t raw_value = getDisplayedPlanet().colonists;
    if (!mp32_t(raw_value).isKnown()) {
        // Unknown
        return mn32_t();
    } else {
        // Known
        return mn32_t(raw_value - cargo_arbiter.getReservedCargo(el_Colonists));
    }
}

void
GPlanet::setColonists(mn32_t clans)
{
    uint32_t old_raw = getDisplayedPlanet().colonists;
    uint32_t new_raw = clans.getRawValue();
    if (clans.isKnown()) {
        new_raw += cargo_arbiter.getReservedCargo(el_Colonists);
    }
    if (old_raw != new_raw) {
        createPlanetInfo().data.colonists = clans.getRawValue();
        markDirty();
    }
}


/*
 *  Native Accessors
 */

mp16_t
GPlanet::getNativeGovernment() const
{
    return mp16_t(getDisplayedPlanet().native_gov);
}

void
GPlanet::setNativeGovernment(mp16_t gov)
{
    createPlanetInfo().data.native_gov = gov.getRawValue();
    markDirty();
}

mn16_t
GPlanet::getNativeHappiness() const
{
    return mn16_t(getDisplayedPlanet().native_happy);
}

void
GPlanet::setNativeHappiness(mn16_t happiness)
{
    createPlanetInfo().data.native_happy = happiness.getRawValue();
    markDirty();
}

mp16_t
GPlanet::getNativeRace() const
{
    return mp16_t(getDisplayedPlanet().native_race);
}

void
GPlanet::setNativeRace(mp16_t race)
{
    createPlanetInfo().data.native_race = race.getRawValue();
    markDirty();
}

mp16_t
GPlanet::getNativeTax() const
{
    return mp16_t(getDisplayedPlanet().native_tax);
}

void
GPlanet::setNativeTax(mp16_t tax)
{
    createPlanetInfo().data.native_tax = tax.getRawValue();
    markDirty();
}

mp32_t
GPlanet::getNatives() const
{
    return mp32_t(getDisplayedPlanet().natives);
}

void
GPlanet::setNatives(mp32_t natives)
{
    createPlanetInfo().data.natives = natives.getRawValue();
    markDirty();
}

bool
GPlanet::isKnownToHaveNatives() const
{
    /* We claim the planet has natives if
       ...we know that it has some from RGA        -or-
       ...we know the race (bioscan)               -or-
       ...we know the population (pillage) */
    return planet_info != 0
        && (planet_info->known_to_have_natives
            || (mp16_t(planet_info->data.native_race).isKnown()
                && planet_info->data.native_race != nr_None)
            || (mp32_t(planet_info->data.natives).isKnown()
                && planet_info->data.natives != 0));
}

void
GPlanet::setKnownToHaveNatives(bool known)
{
    createPlanetInfo().known_to_have_natives = known;
    markDirty();
}


/*
 *  FCode accessors
 */

string_t
GPlanet::getFCode() const
{
    if (planet_info != 0) {
        return convertGameToUtf8(getBasicString(planet_info->data.fcode));
    } else {
        return string_t();
    }
}

void
GPlanet::setFCode(const string_t& str)
{
    storeBasicString(createPlanetInfo().data.fcode, convertUtf8ToGame(str));
    markDirty();
}

bool
GPlanet::isFCodeKnown() const
{
    return planet_info != 0
        && (planet_info->data.fcode[0] != UNKNOWNC
            || planet_info->data.fcode[1] != UNKNOWNC
            || planet_info->data.fcode[2] != UNKNOWNC);
}


/*
 *  Starbase Building Accessors
 */

bool
GPlanet::isBuildingBase() const
{
    return hasFullPlanetData()
        && planet_info->data.build_base;
}

void
GPlanet::setBuildBaseFlag(bool b)
{
    ASSERT(hasFullPlanetData());
    if (b != isBuildingBase()) {
        planet_info->data.build_base = b;
        markDirty();
    }
}


/*
 *  Environment Accessors
 */

mp16_t
GPlanet::getOreDensity(int index) const
{
    return mp16_t(getDisplayedPlanet().ore_density[index]);
}

void
GPlanet::setOreDensity(int index, mp16_t amount)
{
    createPlanetInfo().data.ore_density[index] = amount.getRawValue();
    markDirty();
}

mp32_t
GPlanet::getOreGround(int index) const
{
    return mp32_t(getDisplayedPlanet().ore_ground[index]);
}

void
GPlanet::setOreGround(int index, mp32_t amount)
{
    createPlanetInfo().data.ore_ground[index] = amount.getRawValue();
    markDirty();
}

mp16_t
GPlanet::getTemperature() const
{
    mp16_t temp_code = mp16_t(getDisplayedPlanet().temp_code);
    if (temp_code.isKnown()) {
        return mp16_t(100 - temp_code);
    } else {
        return temp_code;
    }
}

void
GPlanet::setTemperature(mp16_t temp)
{
    if (temp.isKnown()) {
        createPlanetInfo().data.temp_code = 100 - temp;
    } else if (planet_info != 0) {
        planet_info->data.temp_code = mp16_t();
    }
    markDirty();
}


/*
 *  Cargo Accessors
 */

/** Get cargo of a particular type, as to be used by most code.
    This will remove reservations from the amount reported. */
mn32_t
GPlanet::getCargo(GCargoType type) const
{
    mp32_t have = getCargoRaw(type);
    if (have.isKnown()) {
        return mn32_t(have - cargo_arbiter.getReservedCargo(type));
    } else {
        return mn32_t();
    }
}

/** Get cargo, as it is in underlying data structure. */
mp32_t
GPlanet::getCargoRaw(GCargoType type) const
{
    const TPlanet& dpy = getDisplayedPlanet();
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        // Ore
        return mp32_t(dpy.ore_mined[type]);

     case el_Supplies:
        // Supplies
        return mp32_t(dpy.supplies);

     case el_Money:
        // Cash
        return mp32_t(dpy.money);

     case el_Fighters:
        // Fighters
        if (base_info != 0) {
            return mp32_t(base_info->data.ammo_store[10]);
        } else if (map_info.base_kind == NoBase) {
            return mp32_t(0);
        } else {
            return mp32_t();
        }

     case el_Colonists:
        // Colonists
        return mp32_t(dpy.colonists);

     default:
        // Remainder should be torps
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        if (base_info != 0) {
            return mp32_t(base_info->data.ammo_store[type - el_Torps]);
        } else if (map_info.base_kind == NoBase) {
            return mp32_t(0);
        } else {
            return mp32_t();
        }
    }
}

void
GPlanet::changeCargoRaw(GCargoType type, int32 amount)
{
    if (!amount)
        return;

    // FIXME
    ASSERT(hasFullPlanetData());
    markDirty();

    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        planet_info->data.ore_mined[type] += amount;
        return;
     case el_Supplies:
        planet_info->data.supplies += amount;
        return;
     case el_Money:
        planet_info->data.money += amount;
        return;
     case el_Fighters:
        ASSERT(base_info);
        base_info->data.ammo_store[10] += amount;
        return;
     case el_Colonists:
        planet_info->data.colonists += amount;
        return;
     default:
        ASSERT(base_info);
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        base_info->data.ammo_store[type - el_Torps] += amount;
        return;
    }
}

void
GPlanet::setCargoRaw(GCargoType type, mp32_t amount)
{
    // This will create history starbases, so be careful.
    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        createPlanetInfo().data.ore_mined[type] = amount.getRawValue();
        break;
     case el_Supplies:
        createPlanetInfo().data.supplies = amount.getRawValue();
        break;
     case el_Money:
        createPlanetInfo().data.money = amount.getRawValue();
        break;
     case el_Fighters:
        if (base_info != 0 || amount.isKnown()) {
            createBaseInfo().data.ammo_store[10] = amount.getRawValue();
        }
        break;
     case el_Colonists:
        createPlanetInfo().data.colonists += amount.getRawValue();
        break;
     default:
        if (base_info != 0 || amount.isKnown()) {
            createBaseInfo().data.ammo_store[type - el_Torps] += amount.getRawValue();
        }
        break;
    }
}


/*
 *  Simple base Accessors
 */

mp16_t
GPlanet::getBaseDamage() const
{
    return mp16_t(getDisplayedBase().damage);
}

void
GPlanet::setBaseDamage(mp16_t n)
{
    if (base_info != 0 || n.isKnown()) {
        createBaseInfo().data.damage = n.getRawValue();
        markDirty();
    }
}

mp16_t
GPlanet::getBaseMission() const
{
    return mp16_t(getDisplayedBase().mission);
}

void
GPlanet::setBaseMission(mp16_t mission)
{
    if (base_info != 0 || mission.isKnown()) {
        createBaseInfo().data.mission = mission.getRawValue();
        markDirty();
    }
}

mp16_t
GPlanet::getBaseTechLevel(TTechLevel level) const
{
    return mp16_t(getDisplayedBase().tech_levels[level]);
}

void
GPlanet::setBaseTechLevel(TTechLevel level, mp16_t value)
{
    if (base_info != 0 || value.isKnown()) {
        createBaseInfo().data.tech_levels[level] = value.getRawValue();
        markDirty();
    }
}


/*
 *  Shipyard accessors
 */

mp16_t
GPlanet::getBaseShipyardOrder() const
{
    return mp16_t(base_info->data.shipyard_action);
}

mp16_t
GPlanet::getBaseShipyardId() const
{
    return mp16_t(base_info->data.shipyard_id);
}

void
GPlanet::setBaseShipyardOrder(mp16_t action, mp16_t id)
{
    if (base_info != 0 || action.isKnown() || id.isKnown()) {
        TStarbase& b = createBaseInfo().data;
        b.shipyard_action = action.getRawValue();
        b.shipyard_id     = id.getRawValue();
        markDirty();
    }
}


/*
 *  Component storage accessors
 */

mp16_t
GPlanet::getBaseEngineStore(int slot) const
{
    return mp16_t(getDisplayedBase().engine_store[slot-1]);
}

mp16_t
GPlanet::getBaseBeamStore(int slot) const
{
    return mp16_t(getDisplayedBase().beam_store[slot-1]);
}

mp16_t
GPlanet::getBaseHullStoreSlot(int slot) const
{
    return mp16_t(getDisplayedBase().hull_store[slot-1]);
}

mp16_t
GPlanet::getBaseLauncherStore(int slot) const
{
    return mp16_t(getDisplayedBase().launcher_store[slot-1]);
}

/** Get amount of ammunition. This is a convenience function to simplify implementation
    of the script interface.
    \param slot [1,11], where [1,10] are the torps, 11 is fighters */
mp16_t
GPlanet::getBaseAmmoStore(int slot) const
{
    ASSERT(slot > 0 && slot <= NUM_TORPS+1);
    mn32_t tmp;
    if (slot > 0 && slot <= NUM_TORPS) {
        tmp = getCargo(getCargoTypeFromTorpType(slot));
    } else {
        tmp = getCargo(el_Fighters);
    }
    if (tmp.isKnown() && tmp >= 0) {
        return mp16_t(int32_t(tmp));
    } else {
        return mp16_t();
    }
}

/** Set number of engines in starbase storage. */
void
GPlanet::setBaseEngineStore(int slot, mp16_t amount)
{
    ASSERT(slot > 0 && slot <= NUM_ENGINES);
    if (base_info != 0 || amount.isKnown()) {
        createBaseInfo().data.engine_store[slot-1] = amount.getRawValue();
        markDirty();
    }
}

/** Set number of beams in starbase storage. */
void
GPlanet::setBaseBeamStore(int slot, mp16_t amount)
{
    ASSERT(slot > 0 && slot <= NUM_BEAMS);
    if (base_info != 0 || amount.isKnown()) {
        createBaseInfo().data.beam_store[slot-1] = amount.getRawValue();
        markDirty();
    }
}

/** Set number of hulls in starbase storage. Note that this takes a
    truehull slot number, not a real hull Id. */
void
GPlanet::setBaseHullStoreSlot(int slot, mp16_t amount)
{
    ASSERT(slot > 0 && slot <= NUM_HULLS_PER_PLAYER);
    if (base_info != 0 || amount.isKnown()) {
        createBaseInfo().data.hull_store[slot-1] = amount.getRawValue();
        markDirty();
    }
}

/** Set number of torpedo launchers in starbase storage. */
void
GPlanet::setBaseLauncherStore(int slot, mp16_t amount)
{
    ASSERT(slot > 0 && slot <= NUM_TORPS);
    if (base_info != 0 || amount.isKnown()) {
        createBaseInfo().data.launcher_store[slot-1] = amount.getRawValue();
        markDirty();
    }
}


/*
 *  Build order accessors
 */

mp16_t
GPlanet::getBaseBuildHull() const
{
    mp16_t slot = getBaseBuildOrderHullSlot();
    if (slot.isKnown() && base_info != 0 && base_info->data.owner > 0) {
        if (slot > 0) {
            return mp16_t(getTruehull(base_info->data.owner, slot));
        } else {
            return mp16_t(0);
        }
    } else {
        return mp16_t();
    }
}

/** Get base build order. Note that the build order is normalized to contain
    the real hull Id, not the slot number. */
TBuildOrder
GPlanet::getBaseBuildOrder() const
{
    TBuildOrder bo = getDisplayedBase().build;
    bo.hull = getBaseBuildHull();
    return bo;
}

/** Set base build order. Note that the build order is normalized to contain
    the real hull Id, not the slot number. */
void
GPlanet::setBaseBuildOrder(const TBuildOrder& order)
{
    if (base_info != 0 || mp16_t(order.hull).isKnown()) {
        TStarbase& b = createBaseInfo().data;
        if (order.hull == 0) {
            // Blank build order
            b.build.hull       = 0;
            b.build.engine     = 0;
            b.build.beam_type  = 0;
            b.build.beam_count = 0;
            b.build.torp_type  = 0;
            b.build.torp_count = 0;
            b.build.unused     = 0;
        } else if (!mp16_t(order.hull).isKnown() || !mp16_t(b.owner).isKnown() || b.owner == 0) {
            // Impossible
            b.build.hull       = mp16_t().getRawValue();
            b.build.engine     = mp16_t().getRawValue();
            b.build.beam_type  = mp16_t().getRawValue();
            b.build.beam_count = mp16_t().getRawValue();
            b.build.torp_type  = mp16_t().getRawValue();
            b.build.torp_count = mp16_t().getRawValue();
            b.build.unused     = mp16_t().getRawValue();
        } else {
            // Do it
            int slot = getTruehullSlot(b.owner, order.hull);
            ASSERT(slot != 0);
            b.build = order;
            b.build.hull = slot;
        }
        markDirty();
    }
}

/** Get hull slot used for ship being built. Whereas the above functions deal
    with real hull numbers, this one returns a slot; this simplifies things
    every once in a while. */
mp16_t
GPlanet::getBaseBuildOrderHullSlot() const
{
    return mp16_t(getDisplayedBase().build.hull);
}

/** Get ship which is trying to clone at this planet. This runs
    through the ship list, so use with care.
    \param univ universe which contains the ships */
int
GPlanet::findShipCloningHere(const GUniverse& univ) const
{
    // FIXME: check whether 'cln' works
    if (!config.AllowShipCloning())
        return 0;
    for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
        if (univ.getShip(i).isCloningAt(*this)) {
            return i;
        }
    }
    return 0;
}


/*
 *  Build queue accessors
 */
mp16_t
GPlanet::getBaseQueuePosition() const
{
    if (base_info != 0) {
        return base_info->queue_pos;
    } else {
        return mp16_t();
    }
}

void
GPlanet::setBaseQueuePosition(mp16_t pos)
{
    if (base_info != 0 || pos.isKnown()) {
        createBaseInfo().queue_pos = pos;
        markDirty();
    }
}

mp32_t
GPlanet::getBaseQueuePriority() const
{
    if (base_info != 0) {
        return base_info->queue_priority;
    } else {
        return mp32_t();
    }
}

void
GPlanet::setBaseQueuePriority(mp32_t pri)
{
    if (base_info != 0 || pri.isKnown()) {
        createBaseInfo().queue_priority = pri;
        markDirty();
    }
}


/*
 *  Auto build accessors
 */

/** Get autobuild goal for a structure. Known for all planets. */
int
GPlanet::getAutobuildGoal(TPlanetaryStructure ps) const
{
    return map_info.autobuild.goal[ps];
}

/** Set autobuild goal for a structure. */
void
GPlanet::setAutobuildGoal(TPlanetaryStructure ps, int value)
{
    if (value != map_info.autobuild.goal[ps]) {
        map_info.autobuild.goal[ps] = value;
        markDirty();
    }
}

/** Get autobuild speed for a structure. Known for all planets. */
int
GPlanet::getAutobuildSpeed(TPlanetaryStructure ps) const
{
    return map_info.autobuild.speed[ps];
}

/** Set autobuild speed for a structure. */
void
GPlanet::setAutobuildSpeed(TPlanetaryStructure ps, int value)
{
    if (value != map_info.autobuild.speed[ps]) {
        map_info.autobuild.speed[ps] = value;
        markDirty();
    }
}


/*
 *  Data Accesss
 */

/** Create PlanetInfo if it does not exist. Returns reference to it. */
GPlanet::PlanetInfo&
GPlanet::createPlanetInfo()
{
    if (!planet_info) {
        planet_info = new PlanetInfo(map_info.id);
    }
    return *planet_info;
}

/** Create BaseInfo if it does not exist. Returns reference to it. */
GPlanet::BaseInfo&
GPlanet::createBaseInfo()
{
    if (!base_info) {
        base_info = new BaseInfo(map_info.id);
        if (planet_info != 0) {
            base_info->data.owner = planet_info->data.owner;
        }
    }
    return *base_info;
}

/** Get read-only planet data. */
const TPlanet&
GPlanet::getDisplayedPlanet() const
{
    return planet_info != 0 ? planet_info->data : blank_planet;
}

/** Get read-only starbase data. */
const TStarbase&
GPlanet::getDisplayedBase() const
{
    return base_info != 0 ? base_info->data : blank_starbase;
}
