/**
  *  \file game/spec.cc
  *  \brief Specification files.
  *
  *  This file contains the classes which correspond to static game
  *  information (hullspec etc.), and functions to load/access the
  *  static specification files.
  *
  *  \todo Free-form field for starship hulls?
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/spec.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "game/gamedirs.h"
#include "game/host.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/confpars.h"
#include "util/console.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/recode.h"
#include "game/specsn.h"

/** Beam specifications. Indexed 0 to NUM_BEAMS-1 */
static ptr_vector<GBeam> beam_spec;

/** Torpedo specifications. Indexed 0 to NUM_TORPS-1 */
static ptr_vector<GTorpLauncher> torp_spec;

/** Engine specifications. Indexed 0 to NUM_ENGINES-1 */
static ptr_vector<GEngine> eng_spec;

/** Hull specification. Indexed 0 to NNN */
ptr_vector<GHull> hull_spec;

/** Truehull. For each player the hulls he can build. */
TTruehull truehull_spec;

/** \class GComponent
    \brief Component of a game item

    This class represents an abstract component of something
    (weapon, structure, ...) */

GComponent::GComponent(TShortName kind, int id, int mass, int tech, string_t name, GCost cost)
    : kind(kind),
      id(id),
      mass(mass),
      tech(tech),
      name(name),
      shortName(),
      cost(cost)
{ }

GComponent::~GComponent()
{ }

string_t
GComponent::getShortName() const
{
    if (shortName.empty()) {
        return ::getShortName(kind, getId(), getName());
    } else {
        return shortName;
    }
}

/********************************* GBeam *********************************/

/** \class GBeam
    \brief Beam Weapon

    This class represents a beam weapon. */
GBeam::GBeam(int id, const TBeam& data)
    : GWeapon(snBeamShort,
              id,
              data.mass,
              data.tech,
              convertGameToUtf8(getBasicString(data.name)),
              data.cost,
              data.kill,
              data.damage)
{ }

/******************************** GTorpedo *******************************/

/** \class GTorpedo
    \brief Torpedo

    This class represents a torpedo. It differs from GTorpLauncher
    only in the cost it returns. */
GTorpedo::GTorpedo(const GTorpLauncher& launcher)
    : GWeapon(snTorpShort,
              launcher.getId(),
              1,
              launcher.getTech(),
              launcher.getName(),
              GCost(),
              launcher.getKill(),
              launcher.getDamage())
{
    cost.set(el_Money, launcher.getTorpCost());
    cost.set(el_Tritanium, 1);
    cost.set(el_Duranium, 1);
    cost.set(el_Molybdenum, 1);
    setShortName(launcher.getShortName());
}

/***************************** GTorpLauncher *****************************/

/** \class GTorpLauncher
    \brief Torpedo Launcher

    This class represents a torpedo launcher. */
GTorpLauncher::GTorpLauncher(int id, const TTorpedo& data)
    : GWeapon(snTorpShort,
              id,
              data.laun_mass,
              data.tech,
              convertGameToUtf8(getBasicString(data.name)),
              data.laun_cost,
              data.kill,
              data.damage),
      torpCostMC(data.torp_cost)
{ }

GTorpedo
GTorpLauncher::getTorpedo() const
{
    return GTorpedo(*this);
}

/******************************** GFighter *******************************/

/** \class GFighter
    \brief Fighter

    This class represents a fighter. Its main purpose is to provide a
    container for the fighter specs. Unlike the other specs, which
    represent different types which are constant for all players,
    there is only one type of fighters, but that one can differ for
    all players. */

GFighter::GFighter(int id)
    : GWeapon(snFighterShort,
              id,
              1, 1,
              _("Fighter"),
              config.BaseFighterCost(id),
              2, 2)
{
    if (host.isPHost()) {
        kill   = config.FighterBeamKill(id);
        damage = config.FighterBeamExplosive(id);
    }
}

/******************************** GEngine ********************************/

/** \class GEngine
    \brief Engine

    This class represents a starship engine. */
GEngine::GEngine(int id, const TEngine& data)
    : GComponent(snEngineShort,
                 id,
                 0,
                 data.tech,
                 convertGameToUtf8(getBasicString(data.name)),
                 data.cost),
      max_eff_warp(9)
{
    for (std::size_t i = 0; i < countof(fuel_factors); ++i) {
        fuel_factors[i] = data.fuel[i];
    }
    while (max_eff_warp > 1 && getFuelFactor(max_eff_warp) > 120L * max_eff_warp * max_eff_warp) {
        --max_eff_warp;
    }
}

/** Get fuel factor. With the standard fuel consumption formula, this
    is the amount of fuel used by a 100000 kt ship travelling one
    month at the specified speed. The value commonly communicated
    to the user is `getFuelFactor(warp)/(warp*warp)' (fuel per lightyear). */
int
GEngine::getFuelFactor(int warp) const
{
    if (warp == 0)
        return 0;

    ASSERT(warp > 0 && warp < 10);
    return fuel_factors[warp-1];
}

/** Get maximum efficient warp factor. This is default-initialized to
    the warp factor with 120% nominal fuel usage, but can be changed
    using setMaxEfficientWarp(). */
int
GEngine::getMaxEfficientWarp() const
{
    return max_eff_warp;
}

/** Set maximum efficient warp factor.
    \param warp warp factor, 0..9. */
void
GEngine::setMaxEfficientWarp(int warp)
{
    ASSERT(warp >= 0 && warp < 10);
    max_eff_warp = warp;
}

/********************************* GHull *********************************/

// FIXME: can we avoid this being public?
GHull::special_t GHull::racial_abilities;

/** \class GHull
    \brief Starship Hull

    Represents a starship hull. This is information gathered
    together from multiple files: HULLSPEC, HULLFUNC. */
GHull::GHull(int id, const THull& data)
    : GComponent(snHullShort,
                 id,
                 data.mass,
                 data.tech,
                 convertGameToUtf8(getBasicString(data.name)),
                 GCost()),
      external_picture(data.picture),
      /* This remapping conforms to Winplan and PCC 1.x. */
      internal_picture(id == 104 ? 152 : id == 105 ? 153 : data.picture),
      max_fuel(data.fuel),
      max_crew(data.crew),
      num_engines(data.engines),
      max_cargo(data.cargo),
      num_bays(data.bays),
      max_launchers(data.launchers),
      max_beams(data.beams)
{
    cost.set(el_Money, data.mc);
    cost.set(el_Tritanium, data.t);
    cost.set(el_Duranium, data.d);
    cost.set(el_Molybdenum, data.m);
    clearSpecialFunctions();
}

GHull::~GHull()
{ }

/** Modify hull function assignment.
    \param id               Internal (modified) function Id, comes from
                            GHullFunctionData::getIdFromFunction etc.
    \param add              Allow these players to use it
    \param remove           Then disallow these players using it
    \param assign_to_hull   true: this function is available to all ships of this type
                            false: this function is assigned to newly-built ships */
void
GHull::changeSpecialFunction(GHullFunctionData::dev_t id, GPlayerSet add, GPlayerSet remove, bool assign_to_hull)
{
    SpecialAssignment* set = getSpecialFunctionPtr(id, assign_to_hull);
    if (set != 0) {
        set->added += add;
        set->added -= remove;
        set->removed += remove;
        set->removed -= add;
    } else if (!add.empty()) {
        /* Remove-only settings are only relevant for functions that have a variable
           default. Since all to which this applies are already on the list, and thus
           handled by the case above, we need not make a new entry for those. */
        SpecialAssignment n(id, add - remove, remove - add);
        specials[assign_to_hull].push_back(n);
    }
}

/** Enumerate all hull functions related to this hull.
    \param asgn                [out] Functions
    \param player_restriction  [in] List only functions accessible to these players
    \param level_restriction   [in] List only functions accessible at these levels
    \param include_new_ship    [in] Include functions assigned to newly-built ships
    \param include_racial_abilities [in] Include racial abilities. */
void
GHull::enumerateHullFunctions(GHullFunctionList& asgn,
                              GPlayerSet player_restriction,
                              GExpLevelSet level_restriction,
                              bool include_new_ship,
                              bool include_racial_abilities) const
{
    if (include_racial_abilities)
        enumerateSet(racial_abilities, asgn, player_restriction, level_restriction, GHullFunction::AssignedToRace);

    if (include_new_ship)
        enumerateSet(specials[0], asgn, player_restriction, level_restriction, GHullFunction::AssignedToShip);

    enumerateSet(specials[1], asgn, player_restriction, level_restriction, GHullFunction::AssignedToHull);
}

void
GHull::enumerateSet(const special_t& stuff,
                    GHullFunctionList& asgn,
                    GPlayerSet player_restriction,
                    GExpLevelSet level_restriction,
                    GHullFunction::Kind kind) const
{
    for (special_t::size_type i = 0; i < stuff.size(); ++i) {
        GHullFunction func;
        GPlayerSet result;
        if (kind == GHullFunction::AssignedToHull)
            result = GHullFunctionData::instance.getDefaultAssignment(stuff[i].device, *this);
        result += stuff[i].added;
        result -= stuff[i].removed;
        result &= player_restriction;
        if (result.nonempty() && GHullFunctionData::instance.getFunctionDefinition(stuff[i].device, func)) {
            if (func.getLevels().containsAnyOf(level_restriction)) {
                func.setPlayers(result);
                func.setKind(kind);
                asgn.add(func);
            }
        }
    }
}

GPlayerSet
GHull::checkSet(const special_t& stuff,
                int basic_function,
                GExpLevelSet levels,
                bool use_defaults) const
{
    const GHullFunctionData& d = GHullFunctionData::instance;

    GPlayerSet result;
    for (special_t::const_iterator i = stuff.begin(), end = stuff.end(); i != end; ++i) {
        GHullFunction defn;
        GPlayerSet eff;
        if (use_defaults)
            eff = GHullFunctionData::instance.getDefaultAssignment(i->device, *this);
        eff += i->added;
        eff -= i->removed;
        if (eff.nonempty() && d.getFunctionDefinition(i->device, defn))
            if (levels.containsAnyOf(defn.getLevels()))
                if (d.matchBasicFunction(basic_function, defn.getBasicFunctionId()))
                    result |= eff;
    }

    return result;
}

/** Get definition for a hull function. */
GHull::SpecialAssignment*
GHull::getSpecialFunctionPtr(GHullFunctionData::dev_t id, bool assign_to_hull)
{
    special_t& v = specials[assign_to_hull];
    for (std::size_t i = 0; i < v.size(); ++i)
        if (v[i].device == id)
            return &v[i];
    return 0;
}

/** Remove all special functions. */
void
GHull::clearSpecialFunctions()
{
    specials[0].clear();
    specials[1].clear();

    /* Some functions can be modified by config options. Those are described
       by GHullFunctionData::getDefaultAssignment(). We add an entry effectively
       saying "no change", to make enumerateHullFunctions() query
       getDefaultAssignment(). The alternative would be to special-case it.
       Note that this uses the fact that a hf_XXX can be used as a dev_t.
       The converse does not hold. */
    specials[1].push_back(SpecialAssignment(GHullFunctionData::dev_t(hf_Tow),               GPlayerSet(), GPlayerSet()));
    specials[1].push_back(SpecialAssignment(GHullFunctionData::dev_t(hf_Boarding),          GPlayerSet(), GPlayerSet()));
    specials[1].push_back(SpecialAssignment(GHullFunctionData::dev_t(hf_AntiCloakImmunity), GPlayerSet(), GPlayerSet()));
    specials[1].push_back(SpecialAssignment(GHullFunctionData::dev_t(hf_PlanetImmunity),    GPlayerSet(), GPlayerSet()));
    specials[1].push_back(SpecialAssignment(GHullFunctionData::dev_t(hf_FullWeaponry),      GPlayerSet(), GPlayerSet()));
}

/** Check for special function.
    \param basic_function basic function, hf_XXX.
    \param player         check one player only
    \param hull           true: query hull-specific abilities; Every ship of this
                          type <em>owned</em> by the player has the function.
                          false: query ship-specific ability. Every ship of this
                          type <em>built</em> by the player will have the function.
    \param levels         experience level restriction. Check only for a function
                          working at specified levels.
    \returns true iff player can use the function */
bool
GHull::canDoSpecial(int basic_function, int player, bool hull, GExpLevelSet levels) const
{
    return getPlayersThatCan(basic_function, hull, levels).contains(player);
}

/** Get player mask for special function.
    \param basic_function basic function, hf_XXX.
    \param hull           true: query hull-specific abilities;
                          false: ship-specific abilities for newly-built ships
    \param levels         experience level restriction. Return only functions
                          working at specified levels.
    \returns set of all players for whom canDoSpecial() would return true. */
GPlayerSet
GHull::getPlayersThatCan(int basic_function, bool hull, GExpLevelSet levels) const
{
    if (hull)
        return checkSet(specials[true], basic_function, levels, true)
            |  checkSet(racial_abilities, basic_function, levels, false);
    else
        return checkSet(specials[false], basic_function, levels, false);
}

void
GHull::clearRacialAbilities()
{
    racial_abilities.clear();
}

/** Find racial abilities. We define a racial ability to be an ability which the given races
    have one every ship. We'll hide these during normal operation, to avoid cluttering up
    display real-estate with stuff everyone knows.

    This function will identify the abilities, remove them from the individual hulls ir
    possible, and add them to racial_abilities. */
void
GHull::findRacialAbilities()
{
    // sanity check
    if (hull_spec.empty())
        return;

    GHull::special_t& reference = getHull(1).specials[true /* assign-to-hull */];;
    for (GHull::special_t::size_type i = 0; i < reference.size(); ++i) {
        /* Hull #1 has some hull function for a particular set of
           players. Check all other hulls and check who of them has
           that function, too. */
        GHullFunctionData::dev_t dev = reference[i].device;
        GPlayerSet players =
            GHullFunctionData::instance.getDefaultAssignment(dev, getHull(1))
            + reference[i].added
            - reference[i].removed;

        for (int hull_nr = 2; hull_nr <= getNumHulls(); ++hull_nr) {
            GHull::special_t& other = getHull(hull_nr).specials[true];
            bool found = false;
            for (GHull::special_t::size_type j = 0; j < other.size(); ++j) {
                if (other[j].device == dev) {
                    players &= (GHullFunctionData::instance.getDefaultAssignment(dev, getHull(hull_nr))
                                + other[j].added
                                - other[j].removed);
                    found = true;
                    break;
                }
            }
            if (!found || players.empty()) {
                players = GPlayerSet();
                break;
            }
        }

        /* If some have this function on all ships, add it as racial
           ability and stub it out for the ships. We only remove assignments
           that match our racial ability completely, so that the information
           "SSD has planet-immunity for all races" remains available even
           if Klingons/Rebels have that as a racial ability. Otherwise,
           we would list the SSD as immune for everyone but Kli/Reb. The
           disadvantage is that things like "races 1,2,3 have X on all ships,
           except for race 1 on ship Z" should normally better be represented
           as racial abilities for 2 and 3. */
        if (!players.empty()) {
            SpecialAssignment asgn(dev, players, GPlayerSet());
            racial_abilities.push_back(asgn);
            for (int hull_nr = 2; hull_nr <= getNumHulls(); ++hull_nr) {
                GHull::special_t& other = getHull(hull_nr).specials[true];
                for (GHull::special_t::size_type j = 0; j < other.size(); ++j) {
                    if (other[j].device == dev
                        && (GHullFunctionData::instance.getDefaultAssignment(dev, getHull(1))
                            + other[j].added
                            - other[j].removed) == players)
                    {
                        other.erase(other.begin() + j);
                        break;
                    }
                }
            }
        }
    }
}

/******************************** Loading ********************************/

/** Load beamspec file. Throws on error. */
static void
loadBeams()
{
    Ptr<Stream> file = spec_file_dir->openFile("beamspec.dat", Stream::C_READ);

    beam_spec.clear();
    for (int i = 1; i <= NUM_BEAMS; ++i) {
        TBeam data;
        getStructureT(*file, data);
        beam_spec.push_back(new GBeam(i, data));
    }
}

/** Load torpspec file. Throws on error. */
static void
loadTorps()
{
    Ptr<Stream> file = spec_file_dir->openFile("torpspec.dat", Stream::C_READ);

    torp_spec.clear();
    for (int i = 1; i <= NUM_TORPS; ++i) {
        TTorpedo data;
        getStructureT(*file, data);
        torp_spec.push_back(new GTorpLauncher(i, data));
    }
}

/** Load engspec file. Throws on error. */
static void
loadEngines()
{
    Ptr<Stream> file = spec_file_dir->openFile("engspec.dat", Stream::C_READ);

    eng_spec.clear();
    for (int i = 1; i <= NUM_ENGINES; ++i) {
        TEngine data;
        getStructureT(*file, data);
        eng_spec.push_back(new GEngine(i, data));
    }
}

/** Forget all special functions of all hulls. */
static void
clearSpecials()
{
    for (int i = 1; i < getNumHulls(); ++i)
        getHull(i).clearSpecialFunctions();
}

namespace {
    /** HULLFUNC.TXT parser. */
    class HullfuncParser : public ConfigParser {
        int chosen_function;
        bool assign_to_hull;
        GExpLevelSet levels;
        std::vector<bool> chosen_hulls;

        void performAssignments(GHullFunctionData::dev_t function, GPlayerSet add, GPlayerSet remove);

     public:
        HullfuncParser()
            : chosen_function(-1), assign_to_hull(true),
              levels(GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS))
            { }
        void assignSpecial(int line_id, string_t value, bool byrace);
        void assign(int line_id, string_t name, string_t value, string_t line);
        void error(int line_id, string_t msg);
    };
}

void
HullfuncParser::assign(int line_id, string_t name, string_t value, string_t /*line*/)
{
    if (stringMatch("Initialize", name)) {
        /* "Initialize = Clear" or "Initialize = Default" */
        if (stringMatch("Clear", value)) {
            clearSpecials();
        } else if (stringMatch("Default", value)) {
            clearSpecials();
            GHullFunctionData::instance.performDefaultAssignments(hull_spec);
        } else {
            error(line_id, _("Invalid argument to `Initialize'"));
        }
    } else if (stringMatch("Hull", name)) {
        // Initialize hull set
        chosen_hulls.clear();
        chosen_hulls.resize(getNumHulls());

        // parse
        while (value.size() > 0) {
            string_t first = strUCase(strTrim(strFirst(value, ",")));
            if (first.size() > 0) {
                if (first[0] >= '0' && first[1] <= '9') {
                    // Number or range
                    int min = 1, max = getNumHulls();
                    string_t::size_type pos;
                    if (parseRange(first, min, max, pos)
                        || (pos > 0 && parseRange(first.substr(0, pos), min, max, pos)))
                    {
                        if (min > 0 && min <= max && max <= getNumHulls())
                            for (int i = min; i <= max; ++i)
                                chosen_hulls[i-1] = true;
                    } else {
                        // FIXME: print a warning?
                    }
                } else {
                    // Name or '*'
                    if (first == "*") {
                        std::fill(chosen_hulls.begin(), chosen_hulls.end(), true);
                    } else {
                        for (int i = 1; i <= getNumHulls(); ++i) {
                            string_t this_hull_name = strUCase(getHull(i).getName());
                            if (this_hull_name.size() >= first.size()
                                && this_hull_name.compare(0, first.size(), first) == 0)
                            {
                                chosen_hulls[i-1] = true;
                                break;
                            }
                        }
                    }
                }
            }
            if (!strRemove(value, ","))
                break;
        }
    } else if (stringMatch("Function", name)) {
        /* "Function = Number (garbage)" or "Function = Name" */
        int h;
        string_t::size_type pos;
        if ((stringValue(value, h, pos) || (pos > 0 && stringValue(value.substr(0, pos), h)))
            && h >= 0)
        {
            chosen_function = h;
        } else if (const GBasicHullFunction* hf = GHullFunctionData::instance.getBasicFunctionByName(value, true)) {
            chosen_function = hf->getId();
        } else {
            error(line_id, _("Invalid hull function"));
        }
    } else if (stringMatch("Racesallowed", name)) {
        assignSpecial(line_id, value, true);
        levels = GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS);
    } else if (stringMatch("Playersallowed", name)) {
        assignSpecial(line_id, value, false);
        levels = GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS);
    } else if (stringMatch("Assignto", name)) {
        assign_to_hull = !host.hasShipSpecificFunctions() || stringMatch("Hull", value);
    } else if (stringMatch("Level", name)) {
        int min, max;
        string_t::size_type pos;
        if (stringValue(value, min, pos)) {
            // just a number
            max = NUM_EXPERIENCE_LEVELS;
        } else {
            if (!stringValue(value.substr(0, pos), min)) {
                error(line_id, _("Invalid level number"));
                return;
            }
            // what follows behind the number?
            value = strLTrim(value.substr(pos));
            if (!value.size()) {
                max = NUM_EXPERIENCE_LEVELS;
            } else if (value[0] == '-') {
                if (value.size() == 1) {
                    // "1-"
                    max = NUM_EXPERIENCE_LEVELS;
                } else if (stringValue(value.substr(1), max)) {
                    // "1-2", ok
                } else {
                    error(line_id, _("Invalid level range"));
                    return;
                }
            } else {
                error(line_id, _("Invalid level number"));
                return;
            }
        }
        // We now have a range. Is it valid?
        if (min < 0 || max > NUM_EXPERIENCE_LEVELS || min > max)
            error(line_id, _("Invalid level range"));

        levels = GExpLevelSet::allUpTo(max);
        if (min != 0)
            levels -= GExpLevelSet::allUpTo(min-1);
    } else {
        error(line_id, _("Invalid keyword"));
    }
}

void
HullfuncParser::performAssignments(GHullFunctionData::dev_t function, GPlayerSet add, GPlayerSet remove)
{
    for (int i = 1; i <= int(chosen_hulls.size()); ++i)
        if (chosen_hulls[i-1])
            if (i <= getNumHulls())
                getHull(i).changeSpecialFunction(function, add, remove, assign_to_hull);
}

/** Assign special function.
    \param line_id line number
    \param value   list of players
    \param byrace  if true, the list contains race numbers. Otherwise,
                   player numbers. */
void
HullfuncParser::assignSpecial(int line_id, string_t value, bool byrace)
{
    if (chosen_function < 0) {
        error(line_id, _("No function selected for assignment"));
        return;
    }
    if (chosen_hulls.size() <= 0) {
        error(line_id, _("No hull selected for assignment"));
        return;
    }

    GHullFunctionData::dev_t eff_function =
        GHullFunctionData::instance.getIdFromFunction(GHullFunction(chosen_function, levels));

    GPlayerSet players_to_add, players_to_remove;

    if (!host.hasCumulativeHullfunc())
        players_to_remove = GPlayerSet::allUpTo(NUM_PLAYERS);
    while (value.length()) {
        bool neg = false;
        if (value[0] == '*') {
            value.erase(0, 1);
            players_to_remove = GPlayerSet();
            players_to_add = GPlayerSet::allUpTo(NUM_PLAYERS);
        } else if (value[0] == '-') {
            value.erase(0, 1);
            if (value.length() && value[0] >= '0' && value[0] <= '9') {
                neg = true;
                goto digit;
            }
            players_to_add = GPlayerSet();
            players_to_remove = GPlayerSet::allUpTo(NUM_PLAYERS);
        } else if (value[0] == '+') {
            value.erase(0, 1);
            if (value.length() && value[0] >= '0' && value[0] <= '9')
                goto digit;
            players_to_remove = GPlayerSet();
            players_to_add = GPlayerSet::allUpTo(NUM_PLAYERS);
        } else if (value[0] >= '0' && value[0] <= '9') {
         digit:
            string_t::size_type pos;
            int num;
            if (stringValue(value, num, pos)) {
                value.clear();
            } else {
                if (pos && stringValue(value.substr(0, pos), num)) {
                    /* okay */
                    value.erase(0, pos);
                } else {
                    error(line_id, _("Invalid number"));
                    return;
                }
            }
            GPlayerSet set;
            if (byrace) {
                set += getPlayersOfRace(num);
            } else {
                if (num > 0 && num <= NUM_PLAYERS) {
                    set += num;
                } else {
                    error(line_id, _("Invalid player number"));
                    return;
                }
            }
            if (neg)
                players_to_add -= set, players_to_remove += set;
            else
                players_to_add += set, players_to_remove -= set;
        } else {
            value.erase(0, 1);
        }
    }

    performAssignments(eff_function, players_to_add, players_to_remove);
}

void
HullfuncParser::error(int line_id, string_t msg)
{
    console.write(LOG_WARN, format(_("Line %d: %s"), line_id, msg));
}

/** Load Hull function file. */
static void
loadHullfunc(Ptr<Stream> file, bool in_section)
{
    HullfuncParser parser;
    parser.setCharacterSet(getGameCharacterSet());
    parser.configure("hullfunc", in_section);
    parser.parseFile(*file);
}

/** Load hullspec and truehull file. Throws on error.
    \todo Load HULLSPEC.CC(?)
    \pre  configuration has been loaded. */
static void
loadHulls()
{
    Ptr<Stream> file = spec_file_dir->openFile("hullspec.dat", Stream::C_READ);

    hull_spec.erase(hull_spec.begin(), hull_spec.end());
    GHull::clearRacialAbilities();

    int nhulls = file->getSize() / THull::size;
    if (nhulls < 10)
        throw FileTooShortException(*file);
    if (nhulls > MAX_HULLS)
        nhulls = MAX_HULLS;

    for (int i = 1; i <= nhulls; ++i) {
        THull data;
        getStructureT(*file, data);
        hull_spec.push_back(new GHull(i, data));
    }

    file = spec_file_dir->openFile("truehull.dat", Stream::C_READ);

    getStructureT(*file, truehull_spec);

    for (int p = 0; p < NUM_PLAYERS; ++p)
        for (int i = 0; i < NUM_HULLS_PER_PLAYER; ++i)
            if (truehull_spec.hulls[p][i] < 0 || unsigned(truehull_spec.hulls[p][i]) > hull_spec.size())
                throw FileFormatException(*file, _("File is invalid"));

    GHullFunctionData::instance.loadBasicFunctionDefinitions();
    GHullFunctionData::instance.performDefaultAssignments(hull_spec);

    if ((file = spec_file_dir->openFileNT("shiplist.txt", Stream::C_READ)))
        loadHullfunc(file, false);
    else if ((file = spec_file_dir->openFileNT("hullfunc.txt", Stream::C_READ)))
        loadHullfunc(file, true);

    GHull::findRacialAbilities();
}

/**************************** Public Interface ***************************/

/** Get beam. bid = beam number (1..NUM_BEAMS) */
GBeam&
getBeam(int bid)
{
    ASSERT(bid > 0 && bid <= NUM_BEAMS);
    return *beam_spec[bid-1];
}

/** Get torpedo launcher. tid = torpedo number (1..NUM_TORPS) */
GTorpLauncher&
getTorp(int tid)
{
    ASSERT(tid > 0 && tid <= NUM_TORPS);
    return *torp_spec[tid-1];
}

/** Get engine. eid = engine number (1..NUM_ENGINES) */
GEngine&
getEngine(int eid)
{
    ASSERT(eid > 0 && eid <= NUM_ENGINES);
    return *eng_spec[eid-1];
}

/** Get hull. hid = hull number (1..getNumHulls()) */
GHull&
getHull(int hid)
{
    ASSERT(hid > 0 && unsigned(hid) <= hull_spec.size());
    return *hull_spec[hid-1];
}

/** Get truehull slot. pid = player number, slot = 1..20 */
int
getTruehull(int pid, int slot)
{
    ASSERT(pid > 0 && pid <= NUM_PLAYERS);
    ASSERT(slot > 0 && slot <= NUM_HULLS_PER_PLAYER);
    return truehull_spec.hulls[pid-1][slot-1];
}

/** Get Truehull slot. Returns the position of hull hid in player
    pid's Truehull list, [1, NUM_HULLS_PER_PLAYER]. Returns 0 if he can
    not build that hull. */
int
getTruehullSlot(int pid, int hid)
{
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
        if (getTruehull(pid, i) == hid)
            return i;
    return 0;
}

/***************************** Initialisation ****************************/

/** Load Specification Files.
    \pre initGameGlobals() has been called.
    \pre Configuration was loaded. */
static void
initSpecification()
{
    initShortNames();
    loadTorps();
    loadBeams();
    loadEngines();
    loadHulls();
}

const ModuleInitializer mod_spec[] = {
    /* Load specs upon entering NoGame or directory */
    { OnEnterNoGame | OnEnterDirectory, initSpecification },
    { 0, 0 },
};
