/**
  *  \file int/csource.h
  *  \brief Interpreter: Command Sources
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_CSOURCE_H
#define PCC_V2_INT_CSOURCE_H

#include <vector>
#include "int/lex.h"
#include "util/recode.h"

class TextFile;

/** Command source. Provides a stream of script lines, as well as a
    tokenizer, for use by the compiler. Derived classes implement
    readNextLine by calling setNextLine or setEOF. */
class IntCommandSource {
    IntTokenizer tok;
    int line_nr;
    bool eof;

 protected:
    void setNextLine(string_t s);
    void setEOF();

 public:
    IntCommandSource();
    virtual ~IntCommandSource();
    virtual void readNextLine() = 0;
    virtual bool setEncoding(bool utf8, CharacterSet cs) = 0;

    bool isEOF() const;
    int getLineNumber() const;

    IntTokenizer& getTokenizer();
};

/** Memory command source. Provides script lines from a memory buffer. */
class IntMemoryCommandSource : public IntCommandSource {
    std::vector<string_t> lines;
    uint32_t index;
 public:
    IntMemoryCommandSource();
    IntMemoryCommandSource(string_t line);
    void addLine(const string_t& line);
    void readNextLine();
    bool setEncoding(bool utf8, CharacterSet cs);
};

/** File command source. Provides script lines from a text file. */
class IntFileCommandSource : public IntCommandSource {
    TextFile& tf;
 public:
    IntFileCommandSource(TextFile& tf);
    void readNextLine();
    bool setEncoding(bool utf8, CharacterSet cs);
};




/**************************** Inline Functions ***************************/
/** Set EOF. Call this from readNextLine when you reached end of file. */
inline void
IntCommandSource::setEOF()
{
    eof = true;
}

/** Check for EOF. */
inline bool
IntCommandSource::isEOF() const
{
    return eof;
}

/** Get line number. */
inline int
IntCommandSource::getLineNumber() const
{
    return line_nr;
}

/** Get embedded tokenizer. */
inline IntTokenizer&
IntCommandSource::getTokenizer()
{
    return tok;
}

#endif
