/**
  *  \file int/if/vcrif.cc
  *  \brief Interpreter Game Interface: VCRs
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/racename.h"
#include "game/spec.h"
#include "game/vcrdb.h"
#include "int/args.h"
#include "int/if/ifutil.h"
#include "int/if/vcrif.h"
#include "int/interror.h"
#include "int/value.h"
#include "io/io.h"
#include "util/nls.h"
#include "util/recode.h"
#include "int/propenum.h"
#include "game/classicvcr.h"
#include "int/compound.h"

IntValue*
IFVcrGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!haveDisplayedTurn())
        return 0;
    if (!checkIntArg(id, args.getNext(), 1, getDisplayedTurn().getVcrs().getNumBattles()))
        return 0;
    return new IntVcrContext(id);
}

IntContext*
IFVcrMake()
{
    if (!haveDisplayedTurn())
        return 0;

    GVcrDatabase& db = getDisplayedTurn().getVcrs();
    if (db.getNumBattles() == 0)
        return 0;

    return new IntVcrContext(1);
}

/***************************** IntVcrContext *****************************/

namespace { enum { VcrProperty, LeftProperty, RightProperty }; }

static const TIntMapping vcr_mapping[] = {
    { "ALGORITHM",                ivpAlgorithm,     VcrProperty,   thString },
    { "CAPABILITIES",             ivpFlags,         VcrProperty,   thInt },
    { "LEFT",                     ivsNameFull,      LeftProperty,  thString },
    { "LEFT.AUX",                 ivsAuxName,       LeftProperty,  thString },
    { "LEFT.AUX$",                ivsAuxId,         LeftProperty,  thInt },
    { "LEFT.AUX.AMMO",            ivsAuxAmmo,       LeftProperty,  thInt },
    { "LEFT.AUX.COUNT",           ivsAuxCount,      LeftProperty,  thInt },
    { "LEFT.AUX.SHORT",           ivsAuxShort,      LeftProperty,  thString },
    { "LEFT.BEAM",                ivsBeamName,      LeftProperty,  thString },
    { "LEFT.BEAM$",               ivsBeamId,        LeftProperty,  thInt },
    { "LEFT.BEAM.COUNT",          ivsBeamCount,     LeftProperty,  thInt },
    { "LEFT.BEAM.SHORT",          ivsBeamShort,     LeftProperty,  thString },
    { "LEFT.CREW",                ivsCrew,          LeftProperty,  thInt },
    { "LEFT.CREW$",               ivsCrewRaw,       LeftProperty,  thInt },
    { "LEFT.DAMAGE",              ivsDamage,        LeftProperty,  thInt },
    { "LEFT.FIGHTER.BAYS",        ivsFighterBays,   LeftProperty,  thInt },
    { "LEFT.FIGHTER.COUNT",       ivsFighterCount,  LeftProperty,  thInt },
    { "LEFT.HULL",                ivsHullName,      LeftProperty,  thString },
    { "LEFT.HULL$",               ivsHullId,        LeftProperty,  thInt },
    { "LEFT.ID",                  ivsId,            LeftProperty,  thInt },
    { "LEFT.IMAGE",               ivsImage,         LeftProperty,  thInt },
    { "LEFT.LEVEL",               ivsLevel,         LeftProperty,  thInt },
    { "LEFT.MASS",                ivsMass,          LeftProperty,  thInt },
    { "LEFT.NAME",                ivsName,          LeftProperty,  thString },
    { "LEFT.OWNER",               ivsOwnerShort,    LeftProperty,  thString },
    { "LEFT.OWNER$",              ivsOwnerId,       LeftProperty,  thInt },
    { "LEFT.OWNER.ADJ",           ivsOwnerAdj,      LeftProperty,  thString },
    { "LEFT.SHIELD",              ivsShield,        LeftProperty,  thInt },
    { "LEFT.STATUS",              ivsStatus,        LeftProperty,  thString },
    { "LEFT.STATUS$",             ivsStatusRaw,     LeftProperty,  thString },
    { "LEFT.TORP",                ivsTorpName,      LeftProperty,  thString },
    { "LEFT.TORP$",               ivsTorpId,        LeftProperty,  thInt },
    { "LEFT.TORP.COUNT",          ivsTorpCount,     LeftProperty,  thInt },
    { "LEFT.TORP.LCOUNT",         ivsTorpLCount,    LeftProperty,  thInt },
    { "LEFT.TORP.SHORT",          ivsTorpShort,     LeftProperty,  thString },
    { "LEFT.TYPE",                ivsType,          LeftProperty,  thString },
    { "LEFT.TYPE.SHORT",          ivsTypeShort,     LeftProperty,  thString },
    { "MAGIC",                    ivpMagic,         VcrProperty,   thInt },
    { "NUMUNITS",                 ivpNumUnits,      VcrProperty,   thInt },
    { "RIGHT",                    ivsNameFull,      RightProperty, thString },
    { "RIGHT.AUX",                ivsAuxName,       RightProperty, thString },
    { "RIGHT.AUX$",               ivsAuxId,         RightProperty, thInt },
    { "RIGHT.AUX.AMMO",           ivsAuxAmmo,       RightProperty, thInt },
    { "RIGHT.AUX.COUNT",          ivsAuxCount,      RightProperty, thInt },
    { "RIGHT.AUX.SHORT",          ivsAuxShort,      RightProperty, thString },
    { "RIGHT.BEAM",               ivsBeamName,      RightProperty, thString },
    { "RIGHT.BEAM$",              ivsBeamId,        RightProperty, thInt },
    { "RIGHT.BEAM.COUNT",         ivsBeamCount,     RightProperty, thInt },
    { "RIGHT.BEAM.SHORT",         ivsBeamShort,     RightProperty, thString },
    { "RIGHT.CREW",               ivsCrew,          RightProperty, thInt },
    { "RIGHT.CREW$",              ivsCrewRaw,       RightProperty, thInt },
    { "RIGHT.DAMAGE",             ivsDamage,        RightProperty, thInt },
    { "RIGHT.FIGHTER.BAYS",       ivsFighterBays,   RightProperty, thInt },
    { "RIGHT.FIGHTER.COUNT",      ivsFighterCount,  RightProperty, thInt },
    { "RIGHT.HULL",               ivsHullName,      RightProperty, thString },
    { "RIGHT.HULL$",              ivsHullId,        RightProperty, thInt },
    { "RIGHT.ID",                 ivsId,            RightProperty, thInt },
    { "RIGHT.IMAGE",              ivsImage,         RightProperty, thInt },
    { "RIGHT.LEVEL",              ivsLevel,         RightProperty, thInt },
    { "RIGHT.MASS",               ivsMass,          RightProperty, thInt },
    { "RIGHT.NAME",               ivsName,          RightProperty, thString },
    { "RIGHT.OWNER",              ivsOwnerShort,    RightProperty, thString },
    { "RIGHT.OWNER$",             ivsOwnerId,       RightProperty, thInt },
    { "RIGHT.OWNER.ADJ",          ivsOwnerAdj,      RightProperty, thString },
    { "RIGHT.SHIELD",             ivsShield,        RightProperty, thInt },
    { "RIGHT.STATUS",             ivsStatus,        RightProperty, thString },
    { "RIGHT.STATUS$",            ivsStatusRaw,     RightProperty, thString },
    { "RIGHT.TORP",               ivsTorpName,      RightProperty, thString },
    { "RIGHT.TORP$",              ivsTorpId,        RightProperty, thInt },
    { "RIGHT.TORP.COUNT",         ivsTorpCount,     RightProperty, thInt },
    { "RIGHT.TORP.LCOUNT",        ivsTorpLCount,    RightProperty, thInt },
    { "RIGHT.TORP.SHORT",         ivsTorpShort,     RightProperty, thString },
    { "RIGHT.TYPE",               ivsType,          RightProperty, thString },
    { "RIGHT.TYPE.SHORT",         ivsTypeShort,     RightProperty, thString },
    { "SEED",                     ivpSeed,          VcrProperty,   thInt },
    { "TYPE$",                    ivpType,          VcrProperty,   thInt },
    { "UNIT",                     ivpUnits,         VcrProperty,   thArray },
};

IntVcrContext::IntVcrContext(int vcrnr)
    : vcrnr(vcrnr)
{ }

IntVcrContext::~IntVcrContext()
{ }

// IntContext:
int32_t
IntVcrContext::lookup(const string_t& name)
{
    return lookupName(name, vcr_mapping, countof(vcr_mapping));
}

void
IntVcrContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntVcrContext::get(int32_t index)
{
    /* Attempt to fetch VCR entry */
    if (!haveDisplayedTurn())
        return 0;

    GVcrDatabase& db = getDisplayedTurn().getVcrs();
    if (vcrnr <= 0 || vcrnr > db.getNumBattles())
        return 0;

    /* We have a VCR, and we have a valid index, so this will work out */
    GVcrEntry& vcr = db.getBattle(vcrnr-1);
    switch (vcr_mapping[index].domain) {
     case VcrProperty:
        return getVcrProperty(vcr, IntVcrProperty(vcr_mapping[index].index));
     case LeftProperty:
        return getVcrSideProperty(vcr, 0, IntVcrSideProperty(vcr_mapping[index].index));
     case RightProperty:
        return getVcrSideProperty(vcr, 1, IntVcrSideProperty(vcr_mapping[index].index));
     default:
        return 0;
    }
}

bool
IntVcrContext::next()
{
    if (haveDisplayedTurn() && vcrnr < getDisplayedTurn().getVcrs().getNumBattles()) {
        ++vcrnr;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntVcrContext::clone()
{
    return new IntVcrContext(vcrnr);
}

GObject*
IntVcrContext::getObject()
{
    return 0;
}

string_t
IntVcrContext::toString(bool /*readable*/)
{
    return format("Vcr(%d)", vcrnr);
}

void
IntVcrContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notAssignable();
}

void
IntVcrContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(vcr_mapping, countof(vcr_mapping));
}

/********************************* Sides *********************************/

namespace {
    class IntVcrSideArray : public IntIndexableValue {
     public:
        IntVcrSideArray(GVcrEntry& e);

        // IntIndexableValue:
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& args, IntValue* value);
        IntContext* makeFirstContext();

        // IntValue:
        IntVcrSideArray* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
     private:
        GVcrEntry& entry;
    };

    class IntVcrSideContext : public IntContext {
     public:
        IntVcrSideContext(GVcrEntry& entry, int side);
        ~IntVcrSideContext();

        // IntContext:
        int32_t   lookup(const string_t& name);
        void      set(int32_t index, IntValue* value);
        IntValue* get(int32_t index);
        bool      next();
        IntContext* clone();
        GObject*  getObject();
        string_t  toString(bool readable);
        void      store(IntTagNode& sv, Stream& aux);
        void      enumProperties(IntPropertyAcceptor& acceptor);

     private:
        GVcrEntry& entry;
        int side;
    };
}

IntVcrSideArray::IntVcrSideArray(GVcrEntry& e)
    : entry(e)
{ }

IntValue*
IntVcrSideArray::get(IntArgBlock& args)
{
    // Validate args
    int32_t i;
    args.checkArgumentCount(1);
    checkIntArg(i, args.getNext(), 1, entry.getNumObjects());

    // Get result
    return new IntVcrSideContext(entry, i-1);
}

void
IntVcrSideArray::set(IntArgBlock&, IntValue*)
{
    throw IntError::notAssignable();
}

IntContext*
IntVcrSideArray::makeFirstContext()
{
    if (entry.getNumObjects() > 0) {
        return new IntVcrSideContext(entry, 0);
    } else {
        return 0;
    }
}

IntVcrSideArray*
IntVcrSideArray::clone()
{
    return new IntVcrSideArray(entry);
}

string_t
IntVcrSideArray::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntVcrSideArray::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

namespace { enum { SideProperty, SideId }; }

static const TIntMapping side_mapping[] = {
    { "AUX",                 ivsAuxName,       SideProperty,  thString },
    { "AUX$",                ivsAuxId,         SideProperty,  thInt },
    { "AUX.AMMO",            ivsAuxAmmo,       SideProperty,  thInt },
    { "AUX.COUNT",           ivsAuxCount,      SideProperty,  thInt },
    { "AUX.SHORT",           ivsAuxShort,      SideProperty,  thString },
    { "BEAM",                ivsBeamName,      SideProperty,  thString },
    { "BEAM$",               ivsBeamId,        SideProperty,  thInt },
    { "BEAM.COUNT",          ivsBeamCount,     SideProperty,  thInt },
    { "BEAM.SHORT",          ivsBeamShort,     SideProperty,  thString },
    { "CREW",                ivsCrew,          SideProperty,  thInt },
    { "CREW$",               ivsCrewRaw,       SideProperty,  thInt },
    { "DAMAGE",              ivsDamage,        SideProperty,  thInt },
    { "FIGHTER.BAYS",        ivsFighterBays,   SideProperty,  thInt },
    { "FIGHTER.COUNT",       ivsFighterCount,  SideProperty,  thInt },
    { "HULL",                ivsHullName,      SideProperty,  thString },
    { "HULL$",               ivsHullId,        SideProperty,  thInt },
    { "ID",                  ivsId,            SideProperty,  thInt },
    { "IMAGE",               ivsImage,         SideProperty,  thInt },
    { "INDEX",               0,                SideId,        thInt },
    { "ISPLANET",            ivsIsPlanet,      SideProperty,  thBool },
    { "LEVEL",               ivsLevel,         SideProperty,  thInt },
    { "MASS",                ivsMass,          SideProperty,  thInt },
    { "NAME",                ivsName,          SideProperty,  thString },
    { "NAME.FULL",           ivsNameFull,      SideProperty,  thString },
    { "OWNER",               ivsOwnerShort,    SideProperty,  thString },
    { "OWNER$",              ivsOwnerId,       SideProperty,  thInt },
    { "OWNER.ADJ",           ivsOwnerAdj,      SideProperty,  thString },
    { "SHIELD",              ivsShield,        SideProperty,  thInt },
    { "STATUS",              ivsStatus,        SideProperty,  thString },
    { "STATUS$",             ivsStatusRaw,     SideProperty,  thString },
    { "TORP",                ivsTorpName,      SideProperty,  thString },
    { "TORP$",               ivsTorpId,        SideProperty,  thInt },
    { "TORP.COUNT",          ivsTorpCount,     SideProperty,  thInt },
    { "TORP.LCOUNT",         ivsTorpLCount,    SideProperty,  thInt },
    { "TORP.SHORT",          ivsTorpShort,     SideProperty,  thString },
    { "TYPE",                ivsType,          SideProperty,  thString },
    { "TYPE.SHORT",          ivsTypeShort,     SideProperty,  thString },
};


IntVcrSideContext::IntVcrSideContext(GVcrEntry& entry, int side)
    : entry(entry),
      side(side)
{ }

IntVcrSideContext::~IntVcrSideContext()
{ }

int32_t
IntVcrSideContext::lookup(const string_t& name)
{
    return lookupName(name, side_mapping, countof(side_mapping));
}

void
IntVcrSideContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntVcrSideContext::get(int32_t index)
{
    switch (side_mapping[index].domain) {
     case SideProperty:
        return getVcrSideProperty(entry, side, IntVcrSideProperty(side_mapping[index].index));
     case SideId:
        return makeIntValue(side+1);
     default:
        return 0;
    }
}

bool
IntVcrSideContext::next()
{
    if (side+1 < entry.getNumObjects()) {
        ++side;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntVcrSideContext::clone()
{
    return new IntVcrSideContext(entry, side);
}

GObject*
IntVcrSideContext::getObject()
{
    return 0;
}

string_t
IntVcrSideContext::toString(bool /*readable*/)
{
    return "#<vcr-object>";
}

void
IntVcrSideContext::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

void
IntVcrSideContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(side_mapping, countof(side_mapping));
}

/****************************** Name Lookup ******************************/

/** Convert status flags to script value.
    \param status Status value
    \param side Side to query */
static IntValue*
makeVcrResult(GVcrEntry& e, int side)
{
    if (e.getPlayability() != GVcrEntry::IsPlayable) {
        return makeStringValue("Invalid");
    } else {
        /* This originally distinguished between Stalemate, Timeout and Disabled,
           which can be derived from the GClassicVcrEntry interface. Ultimately,
           all three cases are a Stalemate, so we use that. How to detect a
           stalemate? We declare a stalemate if ALL units in a fight survive.
           If anyone dies or gets captured, it is no longer a stalemate. */
        int status = e.getOutcome(side);
        if (status == 0) {
            for (int i = 0, n = e.getNumObjects(); i < n; ++i) {
                if (e.getOutcome(i) != 0) {
                    return makeStringValue("Survived");
                }
            }
            return makeStringValue("Stalemate");
        } else if (status < 0) {
            return makeStringValue("Exploded");
        } else {
            return makeStringValue("Captured");
        }
    }
}

/** Get property of a VCR side. Note that the outcome is a property of
    a VCR side. Therefore, this function may need to invoke the VCR
    player. For this reason, the GVcrEntry cannot be const, because it
    will be updated with the computed result.

    There is no longer any protection against this calling the (costly) VCR.
    Therefore, the server module must make sure to never call this for
    ivsStatus / ivsStatusRaw.

    \param vcr The record.
    \param side Side to query (0=left, 1=right)
    \param ivs Property to get */
IntValue*
getVcrSideProperty(GVcrEntry& vcr, int side, IntVcrSideProperty ivs)
{
    /* Note: this implements the owner properties directly instead of relying on
       getPlayerProperty() like most others. This makes it possible to use VCR
       properties without having a turn loaded. */
    /* FIXME: this does not implement PlanetsHaveTubes; for planets with tubes,
       it always returns the fighters. */
    const GVcrObject& o = vcr.getObject(side, false);
    switch (ivs) {
     case ivsAuxAmmo:
        if (o.getBayCount() > 0) {
            return makeIntValue(o.getFighterCount());
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpCount());
        } else {
            return makeIntValue(0);
        }
     case ivsAuxCount:
        if (o.getBayCount() > 0) {
            return makeIntValue(o.getBayCount());
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpLauncherCount());
        } else {
            return 0;
        }
     case ivsAuxId:
        if (o.getBayCount() > 0) {
            return makeIntValue(11);
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpType());
        } else {
            return 0;
        }
     case ivsAuxName:
        if (o.getBayCount() > 0) {
            return makeStringValue("Fighters");
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.getTorpType()).getName());
        } else {
            return 0;
        }
     case ivsAuxShort:
        if (o.getBayCount() > 0) {
            return makeStringValue("Ftr");
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.getTorpType()).getShortName());
        } else {
            return 0;
        }
     case ivsFighterBays:
        return makeIntValue(o.getBayCount());
     case ivsFighterCount:
        return makeIntValue(o.getBayCount() > 0 ? o.getFighterCount() : 0);
     case ivsTorpId:
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpType());
        } else {
            return 0;
        }
     case ivsTorpCount:
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpCount());
        } else {
            return makeIntValue(0);
        }
        
     case ivsTorpLCount:
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpLauncherCount());
        } else {
            return makeIntValue(0);
        }
     case ivsTorpShort:
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.getTorpType()).getShortName());
        } else {
            return 0;
        }
     case ivsTorpName:
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.getTorpType()).getName());
        } else {
            return 0;
        }
     case ivsBeamCount:
        return makeIntValue(o.getBeamCount());
     case ivsBeamId:
        return makeIntValue(o.getBeamType());
     case ivsBeamName:
        if (o.getBeamType() > 0 && o.getBeamType() <= NUM_BEAMS) {
            return makeStringValue(::getBeam(o.getBeamType()).getName());
        } else {
            return 0;
        }
     case ivsBeamShort:
        if (o.getBeamType() > 0 && o.getBeamType() <= NUM_BEAMS) {
            return makeStringValue(::getBeam(o.getBeamType()).getShortName());
        } else {
            return 0;
        }
     case ivsCrew:
        if (o.isPlanet()) {
            return 0;
        } else {
            return makeIntValue(o.getCrew());
        }
     case ivsCrewRaw:
        return makeIntValue(o.getCrew());
     case ivsDamage:
        return makeIntValue(o.getDamage());
     case ivsId:
        return makeIntValue(o.getId());
     case ivsMass:
        return makeIntValue(o.getMass());
     case ivsName:
        return makeStringValue(o.getName());
     case ivsNameFull:
        return makeStringValue(format(o.isPlanet()
                                      ? _("%s (Planet #%d)")
                                      : _("%s (Ship #%d)"),
                                      o.getName(),
                                      o.getId()));
     case ivsOwnerAdj:
        if (o.getOwner() > 0 && o.getOwner() <= NUM_OWNERS) {
            return makeStringValue(player_racenames.getAdjName(o.getOwner()));
        } else {
            return 0;
        }
     case ivsOwnerId:
        return makeIntValue(o.getOwner());
     case ivsOwnerShort:
        if (o.getOwner() > 0 && o.getOwner() <= NUM_OWNERS) {
            return makeStringValue(player_racenames.getShortName(o.getOwner()));
        } else {
            return 0;
        }
     case ivsShield:
        return makeIntValue(o.getShield());
     case ivsType:
     case ivsTypeShort:
        {
            const char* kind;
            if (o.isPlanet()) {
                kind = "Planet";
            } else if (o.getBayCount() > 0) {
                kind = "Carrier";
            } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
                kind = "Torpedo Ship";
            } else if (o.getBeamCount() > 0) {
                kind = "Beam Weapons";
            } else {
                kind = "Freighter";
            }
            if (ivs == ivsTypeShort) {
                return makeStringValue(string_t(kind, 1));
            } else {
                return makeStringValue(kind);
            }
        }
        break;
     case ivsHullName:
     case ivsHullId:
        if (int h = o.getGuessedHull()) {
            if (ivs == ivsHullId) {
                return makeIntValue(h);
            } else {
                return makeStringValue(::getHull(h).getName());
            }
        } else {
            return 0;
        }
     case ivsImage:
        return makeIntValue(o.getGuessedShipPicture());
     case ivsLevel:
        return makeIntValue(o.getExperienceLevel());
     case ivsStatus:
        return makeVcrResult(vcr, side);
     case ivsStatusRaw:
        return makeIntValue(vcr.getOutcome(side));
     case ivsIsPlanet:
        return makeBoolValue(o.isPlanet());
    }
    return 0;
}

/** Get property of a VCR record.
    \param vcr The record
    \param ivp Property to get */
IntValue*
getVcrProperty(GVcrEntry& vcr, IntVcrProperty ivp)
{
    GClassicVcrEntry* ce = dynamic_cast<GClassicVcrEntry*>(&vcr);
    switch (ivp) {
     case ivpSeed:
        if (ce) {
            return makeIntValue(ce->getVcr().getSeed());
        } else {
            return 0;
        }
     case ivpMagic:
        if (ce) {
            return makeIntValue(ce->getVcr().getSignature());
        } else {
            return 0;
        }
     case ivpType:
        if (ce) {
            return makeIntValue(ce->getVcr()(1).isPlanet());
        } else {
            return 0;
        }
     case ivpAlgorithm:
        return makeStringValue(vcr.getAlgorithmName());
     case ivpFlags:
        if (ce) {
            return makeIntValue(ce->getCapabilities());
        } else {
            return 0;
        }
     case ivpNumUnits:
        return makeIntValue(vcr.getNumObjects());
     case ivpUnits:
        return new IntVcrSideArray(vcr);
    }
    return 0;
}
