/**
  *  \file int/keymap.cc
  *  \brief Interpreter: Keymaps
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "defs.h"
#include "int/keymap.h"
#include "int/interror.h"
#include "util/misc.h"
#include "cpluslib/ptrvector.h"
#ifdef HAVE_SDL
#include "ui/event.h"
#endif
#include "cpluslib/misc.h"
#include "cpluslib/format.h"
#include "util/recode.h"
#include "cpluslib/utf8.h"

namespace {
    typedef ptr_vector<IntKeymap> KeymapVector;
    static KeymapVector keymaps;
}

/** Constructor.
    \param name Name of this keymap */
IntKeymap::IntKeymap(const string_t& name)
    : name(name)
{ }

/** Destructor. */
IntKeymap::~IntKeymap()
{ }

/** Look up a key, recursively.
    \param key [in] Key to find
    \param where [out,optional] Keymap in which key is bound
    \return found Entry, if any. Otherwise null. */
const IntKeymap::Entry*
IntKeymap::lookup(const uint32_t key, IntKeymapRef*const where) const
{
    for (std::vector<Entry>::const_iterator it = keys.begin(); it != keys.end(); ++it) {
        if (it->key == key) {
            if (where)
                *where = const_cast<IntKeymap*>(this);
            return &*it;
        }
    }
    for (std::vector<IntKeymap*>::const_iterator it = parents.begin(); it != parents.end(); ++it) {
        if (const Entry* e = (*it)->lookup(key, where)) {
            return e;
        }
    }
    return 0;
}

/** Add key to this keymap.
    \param key         Key to add
    \param command     Command (=atom)
    \param condition   Condition (=atom) */
void
IntKeymap::addKey(const uint32_t key, const uint32_t command, const uint32_t condition)
{
    /* Replace existing key? */
    for (std::vector<Entry>::iterator it = keys.begin(); it != keys.end(); ++it) {
        if (it->key == key) {
            it->command = command;
            it->condition = condition;
            return;
        }
    }

    /* Add new key */
    keys.push_back(Entry(key, command, condition));
}

/** Add parent to this keymap.
    \param km Other keymap
    \throw IntError if km is already a parent of this one */
void
IntKeymap::addParent(IntKeymapRef km)
{
    if (hasParent(km))
        throw IntError("Duplicate parent keymap");
    if (km->hasParent(this))
        throw IntError("Invalid parent keymap");
    parents.push_back(km);
}

/** Given a key, look up its command.
    \param key Key to look up
    \return found command (atom), 0 if none */
uint32_t
IntKeymap::lookupCommand(uint32_t key) const
{
    if (const Entry* e = lookup(key, 0))
        return e->command;
    return 0;
}

/** Given a key, look up its command and place of definition.
    \param key   [in] Key to look up
    \param where [out] Keymap in which this command is bound
    \return found command (atom), 0 if none */
uint32_t
IntKeymap::lookupCommand(uint32_t key, IntKeymapRef& where) const
{
    where = 0;
    if (const Entry* e = lookup(key, &where))
        return e->command;
    return 0;
}

/** Given a key, look up its condition.
    \param key Key to look up
    \return found condition (atom), 0 if none */
uint32_t
IntKeymap::lookupCondition(uint32_t key) const
{
    if (const Entry* e = lookup(key, 0))
        return e->condition;
    return 0;
}

/** Check for parent relation ship, recursively.
    \param km Keymap to find
    \return true iff km is equal to this, or a parent, grandparent, etc., recursively */
bool
IntKeymap::hasParent(IntKeymap* km) const
{
    if (km == this)
        return true;
    for (std::vector<IntKeymap*>::const_iterator it = parents.begin(); it != parents.end(); ++it)
        if ((*it)->hasParent(km))
            return true;
    return false;
}

/** Get keymap by name. Note that the name must be in upper-case, because that's how it
    comes in from the script interface.
    \param name Name to find
    \return Found keymap, 0 if it does not exist. */
IntKeymapRef
IntKeymap::getKeymapByName(string_t name)
{
    for (KeymapVector::iterator it = keymaps.begin(); it != keymaps.end(); ++it)
        if ((*it)->getName() == name)
            return *it;
    return 0;
}

/** Create keymap. Note that the name must be in upper-case, because that's how it
    comes in from the script interface.
    \param name Name to create
    \return New keymap.
    \throw IntError if keymap by this name already exists */
IntKeymapRef
IntKeymap::createKeymap(string_t name)
{
    if (getKeymapByName(name))
        throw IntError("Keymap already exists");

    IntKeymap* km = new IntKeymap(name);
    keymaps.push_back(km);
    return km;
}

/***************************** IntKeymapValue ****************************/

/** Constructor.
    \param keymap Referenced keymap */
IntKeymapValue::IntKeymapValue(IntKeymapRef keymap)
    : keymap(keymap)
{ }

IntKeymapValue::~IntKeymapValue()
{ }

string_t
IntKeymapValue::toString(bool /*readable*/)
{
    return "#<keymap:" + keymap->getName() + ">";
}

void
IntKeymapValue::store(IntTagNode& /*tag*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

IntKeymapValue*
IntKeymapValue::clone()
{
    return new IntKeymapValue(keymap);
}

/******************************* Main Entry ******************************/

/** Initialize Keymap module. */
void
initKeymaps()
{
    doneKeymaps();
}

/** Shutdown Keymap module. */
void
doneKeymaps()
{
    KeymapVector().swap(keymaps);
}

/****************************** Key parsing ******************************/

#ifdef HAVE_SDL
namespace {
    struct KeyPair {
        const char* name;
        uint32_t    value;
    };
}

static const KeyPair key_mods[] = {
    { "CTRL-",  ss_Ctrl },
    { "C-",     ss_Ctrl },
    { "SHIFT-", ss_Shift },
    { "S-",     ss_Shift },
    { "ALT-",   ss_Alt },
    { "A-",     ss_Alt },
    { "META-",  ss_Meta },
    { "M-",     ss_Meta },
};

static const KeyPair key_syms[] = {
    { "BS",        SDLK_BACKSPACE },
    { "BACKSPACE", SDLK_BACKSPACE },
    { "PAUSE",     SDLK_PAUSE },
    { "DBLCLICK",  vk_DoubleClickLocation },
    { "DEL",       SDLK_DELETE },
    { "DOWN",      SDLK_DOWN },
    { "END",       SDLK_END },
    { "ESC",       SDLK_ESCAPE },
    { "F1",        SDLK_F1 },
    { "F2",        SDLK_F2 },
    { "F3",        SDLK_F3 },
    { "F4",        SDLK_F4 },
    { "F5",        SDLK_F5 },
    { "F6",        SDLK_F6 },
    { "F7",        SDLK_F7 },
    { "F8",        SDLK_F8 },
    { "F9",        SDLK_F9 },
    { "F10",       SDLK_F10 },
    { "F11",       SDLK_F11 },
    { "F12",       SDLK_F12 },
    { "F13",       SDLK_F13 },
    { "F14",       SDLK_F14 },
    { "F15",       SDLK_F15 },
    { "HOME",      SDLK_HOME },
    { "INS",       SDLK_INSERT },
    { "LEFT",      SDLK_LEFT },
    // Some numpad keys, including Enter, are mapped away by ui/event.cc:
    // { "LFD",       SDLK_KP_ENTER },
    // { "NUMRET",    SDLK_KP_ENTER },
    // These numpad keys are not available on all systems:
    { "NUM*",      SDLK_KP_MULTIPLY },
    { "NUM+",      SDLK_KP_PLUS },
    { "NUM-",      SDLK_KP_MINUS },
    { "NUM/",      SDLK_KP_DIVIDE },
    { "NUM5",      SDLK_KP5 },
    { "PGDN",      SDLK_PAGEDOWN },
    { "PGUP",      SDLK_PAGEUP },
    { "PRINT",     SDLK_PRINT },
    { "QUIT",      vk_Quit },
    { "RET",       SDLK_RETURN },
    { "ENTER",     SDLK_RETURN },
    { "RIGHT",     SDLK_RIGHT },
    { "SPACE",     ' ' },
    { "SPC",       ' ' },
    { "TAB",       SDLK_TAB },
    { "UP",        SDLK_UP },
    { "WHEELDN",   vk_WheelDown },
    { "WHEELUP",   vk_WheelUp },
};

/** Parse string into key.
    \param str String given by user
    \retval Parsed key, or 0 if invalid string */
uint32_t
parseKey(string_t str)
{
    // Upcase it
    str = strUCase(str);

    // Parse modifiers
    uint32_t result = 0;
    while (1) {
        bool did = false;
        for (uint_t i = 0; i < countof(key_mods); ++i) {
            std::size_t len = std::strlen(key_mods[i].name);
            if (str.compare(0, len, key_mods[i].name, len) == 0) {
                did = true;
                result |= key_mods[i].value;
                str.erase(0, len);
            }
        }
        if (!did)
            break;
    }

    // Parse key
    if (lengthUtf8(str) == 1) {
        uint32_t uc = charAtUtf8(str, 0);
        if (uc >= 'A' && uc <= 'Z') {
            if (result & ss_Shift) {
                // Shifted alphabetical. Because this already is
                // upper-case, just remove the shift.
                result &= ~ss_Shift;
            } else {
                // Non-shifted. Make it lower-case.
                uc += 'a' - 'A';
            }
        }

        if (uc >= 128) {
            // International. We can only bind keys from the game character set.
            uc = uint8_t(getGameCharacterSet().getCharacter(uc));
            if (!uc)
                return 0;
        }

        result |= uc;
    } else {
        bool found = false;
        for (uint_t i = 0; i < countof(key_syms); ++i) {
            if (str == key_syms[i].name) {
                result |= key_syms[i].value;
                found = true;
                break;
            }
        }
        if (!found && str.size() > 2 && str[0] == '#' && str[1] == '$' && str.size() <= 6) {
            // Up to four hex digits
            const char* p = str.c_str()+2;
            char* pout;
            result |= strtoul(p, &pout, 16);
            if (*pout == 0)
                found = true;
        }
        if (!found) {
            return 0;
        }
    }

    return result;
}

/** Format key code into string.
    \param key Keycode
    \retval Key string so that parseKey(return) == key */
string_t
formatKey(uint32_t key)
{
    string_t result;

    // Modifiers
    for (uint_t i = 0; i < countof(key_mods); ++i) {
        if (key & key_mods[i].value) {
            result += key_mods[i].name;
            key &= ~key_mods[i].value;
        }
    }

    // Key
    bool did = false;
    for (uint_t i = 0; i < countof(key_syms); ++i) {
        if (key == key_syms[i].value) {
            result += key_syms[i].name;
            did = true;
            break;
        }
    }

    if (!did) {
        if (key == vk_Unicode) {
            result += "<international>";
        } else if (key >= 'A' && key <= 'Z') {
            // Upper-case letter
            result += "SHIFT-";
            result += char(key);
        } else if (key >= 'a' && key <= 'z') {
            // Lower-case letter
            result += char(key - ('a'-'A'));
        } else if (key >= ' ' && key < 256 && key != 127) {
            // Key from game character set
            appendUtf8(result, getGameCharacterSet().getUnicode(key));
        } else {
            result += format("#$%04X", key);
        }
    }

    return result;
}
#else
/*
 *  If we don't have SDL, use only dummy keymap codes. However, return
 *  a nonzero value so that callers don't fail.
 */
uint32_t
parseKey(string_t /*str*/)
{
    return 1;
}

string_t
formatKey(uint32_t key)
{
    return format("#$%04X", key);
}
#endif
